"""Test neurots.generate.soma code."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=missing-function-docstring
# pylint: disable=protected-access
import json
import os

import morphio
import numpy as np
import pytest
from mock import patch
from numpy.testing import assert_allclose
from numpy.testing import assert_almost_equal
from numpy.testing import assert_array_almost_equal
from numpy.testing import assert_array_equal
from numpy.testing import assert_equal

from neurots import NeuronGrower
from neurots.generate import soma as tested
from neurots.utils import NeuroTSError

_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")


def test_constructors():

    inputs = [
        [np.array([1.0, 2.0, 3.0]), 2.0],
        [np.array([1, 2, 3]), 2],
        [[1.0, 2.0, 3.0], 2.0],
        [[1, 2, 3], 2.0],
        [(1.0, 2.0, 3.0), 2],
        [(1, 2, 3), 2.0],
    ]

    for center, radius in inputs:

        soma = tested.Soma(center, radius)

        assert_equal(len(soma.points), 0)

        assert_equal(type(soma.radius), float)
        assert_almost_equal(soma.radius, 2.0)

        assert_allclose(soma.center, [1.0, 2.0, 3.0])
        assert_equal(type(soma.center), np.ndarray)

    soma = tested.Soma((1.0, 2.0, 3.0), 2.0, points=None)
    assert len(soma.points) == 0

    soma = tested.Soma((1.0, 2.0, 3.0), 2.0, points=[[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])
    assert_allclose(soma.points, [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])

    soma = tested.Soma((1.0, 2.0, 3.0), 2.0)
    rng = np.random.default_rng()
    context = "some_context"
    soma_grower = tested.SomaGrower(soma=soma, context=context, rng=rng)

    assert soma_grower.soma is soma
    assert soma_grower.context == "some_context"
    assert soma_grower._rng is rng


def test_soma_point_from_trunk_direction():

    soma = tested.Soma((1.0, 2.0, 3.0), 2.0)

    phi = 4.5
    theta = 1.2

    point = soma.point_from_trunk_direction(phi, theta)

    expected = np.array([-0.3929398485, -1.822192555, 0.724715509]) + soma.center

    assert_array_almost_equal(point, expected)


def test_soma_orientation_from_point():

    soma = tested.Soma((1.0, 2.0, 3.0), 2.0)

    point = np.array([0.32764176, 0.53033757, 0.74690682])

    vector = point - soma.center

    expected_orientation = vector / np.linalg.norm(vector)

    result = soma.orientation_from_point(point)
    assert_array_almost_equal(expected_orientation, result)


def test_soma_orientation_from_point_exception():

    soma = tested.Soma((1.0, 2.0, 3.0), 2.0)

    point = np.array([1.0, 2.0, 3.0])

    with pytest.raises(ValueError):
        soma.orientation_from_point(point)


def test_soma_contour_point():

    soma = tested.Soma((1.0, 2.0, 3.0), 2.0)

    point = [4.0, 5.0, 6.0]

    expected = [4.0, 5.0, 3.0]

    result = soma.contour_point(point)

    assert_array_almost_equal(expected, result)


def test_soma_grower_one_point_soma():

    soma = tested.Soma((1.0, 2.0, 3.0), 2.0)
    soma_grower = tested.SomaGrower(soma)

    soma_points, soma_diameters = soma_grower._one_point_soma()

    assert_array_equal(soma_points, [soma.center])
    assert_array_equal(soma_diameters, [4.0])


def test_soma_grower_contour_soma():

    soma = tested.Soma((1.0, 2.0, 3.0), 2.0)
    soma_grower = tested.SomaGrower(soma)

    soma.points = [(3.0, 4.0, 1000000.0)]

    expected_pts = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 3.0]])

    # bypass interpolate
    def bypassing_interpolate(points, interpolation=10):
        # pylint: disable=unused-argument
        return expected_pts

    soma_grower.interpolate = bypassing_interpolate

    soma_pts, diameters = soma_grower._contour_soma()

    assert_array_almost_equal(soma_pts, expected_pts)
    assert_allclose(diameters, 0.0)


def test_soma_grower_original_soma():

    soma = tested.Soma((1.0, 2.0, 3.0), 2.0)
    soma_grower = tested.SomaGrower(soma)

    points, diameters = soma_grower._original_soma()

    assert_equal(len(points), 0)
    assert_equal(len(diameters), 0)


def test_soma_grower_interpolate():

    soma = tested.Soma((0.0, 0.0, 0.0), 6.0)
    soma_grower = tested.SomaGrower(soma)

    soma.center = np.asarray([0.0, 0.0, 0.0], dtype=float)
    soma.points = [[0, 0, 0], [1, 0, 0], [0, 1, 1], [1, 0, 0]]

    np.random.seed(0)

    result = soma_grower.interpolate(soma.points, interpolation=4)
    assert_array_equal(
        result,
        [[0, 0, 0], [1, 0, 0], [0, 1, 1]],
    )

    result = soma_grower.interpolate(soma.points, interpolation=7)

    expected = [
        [-1.8115102245941463, -5.720002684106964, 0.0],
        [1.0, 0.0, 0.0],
        [0.0, 1.0, 1.0],
        [-5.857053738900479, -1.3018915083953513, 0.0],
        [-3.6104565197299623, -4.792139785016648, 0.0],
    ]

    assert_array_almost_equal(result, expected)


def test_soma_grower_interpolate_from_neuron():
    np.random.seed(0)

    g = NeuronGrower(
        {
            "origin": [0, 0, 0],
            "grow_types": [],
            "diameter_params": {"method": "default"},
        },
        {
            "soma": {"size": {"norm": {"mean": 6, "std": 0}}},
            "diameter": {"method": "default"},
        },
    )

    g.soma_grower.soma.points = [[0, 0, 0], [1, 0, 0], [0, 1, 1], [1, 0, 0]]

    assert_array_equal(
        g.soma_grower.interpolate(g.soma_grower.soma.points, interpolation=4),
        [[0, 0, 0], [1, 0, 0], [0, 1, 1]],
    )

    result = g.soma_grower.interpolate(g.soma_grower.soma.points, interpolation=7)

    expected = [
        [-1.8115102245941463, -5.720002684106964, 0.0],
        [1.0, 0.0, 0.0],
        [0.0, 1.0, 1.0],
        [-5.857053738900479, -1.3018915083953513, 0.0],
        [-3.6104565197299623, -4.792139785016648, 0.0],
    ]

    assert_array_almost_equal(result, expected)


def test_soma_interpolate_from_neuron_2():
    np.random.seed(0)
    g = NeuronGrower(
        {
            "origin": [0, 0, 0],
            "grow_types": [],
            "diameter_params": {"method": "default"},
        },
        {
            "soma": {"size": {"norm": {"mean": 6, "std": 3}}},
            "diameter": {"method": "default"},
        },
    )

    soma = g.soma_grower.soma
    soma.points = [[0, 0, 0], [1, 0, 0], [0, 1, 1], [1, 0, 0]]

    assert_array_equal(
        g.soma_grower.interpolate(soma.points, interpolation=4),
        [[0, 0, 0], [1, 0, 0], [0, 1, 1]],
    )

    assert_array_almost_equal(
        g.soma_grower.interpolate(soma.points, interpolation=7),
        [
            [0.0, 1.0, 1.0],
            [-6.794973666551906, -9.018932499998446, 0.0],
            [-3.1424513796161775, -10.846096528033044, 0.0],
            [5.211388179030298, -10.017696532443242, 0.0],
            [1.0, 0.0, 0.0],
        ],
    )


def test_soma_interpolate_exception():

    soma = tested.Soma((0.0, 0.0, 0.0), 6.0)
    soma_grower = tested.SomaGrower(soma)

    soma.center = np.asarray([0.0, 0.0, 0.0], dtype=float)

    soma.points = [[1, 1, 1], [2, 2, 2], [3, 3, 3]]

    with pytest.raises(NeuroTSError):
        soma_grower.interpolate(soma.points, interpolation=1)


def test_add_points_from_trunk_angles():

    soma = tested.Soma((1.0, 2.0, 3.0), 2.0)
    soma_grower = tested.SomaGrower(soma)

    trunk_angle_deviations = [0.1, 0.1, 0.1]

    z_angles = [1.0, 1.0, 1.0]

    points = soma_grower.add_points_from_trunk_angles(trunk_angle_deviations, z_angles)

    expected = np.array(
        [
            [0.017229, 3.366182, 4.080605],
            [0.308237, 0.465804, 4.080605],
            [2.674534, 2.168014, 4.080605],
        ]
    )

    assert_array_almost_equal(points, expected)
    assert_array_almost_equal(soma.points, expected)

    # Test with interval [1, 4]
    interval = [1, 4]
    points = soma_grower.add_points_from_trunk_angles(trunk_angle_deviations, z_angles, interval)

    expected = np.array(
        [
            [0.1503733, 3.45273127, 4.08060461],
            [-0.68148648, 2.06997784, 4.08060461],
            [0.03260466, 0.62288711, 4.08060461],
        ]
    )


def test_add_points_from_trunk_absolute_orientation():

    soma = tested.Soma((1.0, 2.0, 3.0), 2.0)

    # Test with trivial orientation
    soma_grower = tested.SomaGrower(soma)

    orientation = np.array([0, 1, 0])
    trunk_absolute_angles = np.array([0.75, 1.0, 0.75])
    z_angles = np.array([1.48090896530732, 1.7240863537476048, 1.2747004358482887])

    points = soma_grower.add_points_from_trunk_absolute_orientation(
        orientation, trunk_absolute_angles, z_angles
    )

    expected = np.array(
        [
            [2.457469863016131, 3.3577737648609998, 3.179532731987588],
            [2.399695853613939, 3.3039517022251426, 3.5835764522666245],
            [2.067933519838869, 3.6632079132149533, 2.694619197355619],
        ]
    )

    assert_array_almost_equal(points, expected)
    assert_array_almost_equal(soma.points, expected)

    soma = tested.Soma((1.0, 2.0, 3.0), 2.0)

    # Test with non trivial orientation
    soma_grower = tested.SomaGrower(soma)

    base_orientation = [0.25, 0.5, 0.75]
    orientation = np.array(base_orientation) / np.linalg.norm(base_orientation)
    trunk_absolute_angles = [0.75, 1.0, 1.0]
    z_angles = [1.806078056398836, 0.947826003221246, 1.512974272395405]

    points = soma_grower.add_points_from_trunk_absolute_orientation(
        orientation, trunk_absolute_angles, z_angles
    )

    expected = np.array(
        [
            [2.473567419532686, 2.4338840563147373, 4.280759027205216],
            [1.0301730699570344, 2.0179373519813315, 4.999691935587396],
            [1.9460185698129842, 2.5623911684086043, 4.6699595921054495],
        ]
    )

    assert_array_almost_equal(points, expected)
    assert_array_almost_equal(soma.points, expected)


def test_add_points_from_orientations():

    soma = tested.Soma((1.0, 2.0, 3.0), 2.0)
    soma_grower = tested.SomaGrower(soma)

    vectors = np.array(
        [
            [0.87112168, 0.45651245, 0.42912960],
            [0.43898550, 0.16391644, 0.03717331],
            [0.42663795, 0.79006525, 0.85176434],
            [0.36492627, 0.54107164, 0.83189980],
        ]
    )

    vectors /= np.linalg.norm(vectors, axis=1)[:, np.newaxis]

    points = soma_grower.add_points_from_orientations(vectors)

    expected = [soma.center + soma.radius * (v / np.linalg.norm(v)) for v in vectors]

    assert_array_almost_equal(points, expected)
    assert_array_almost_equal(soma.points, expected)


def test_build():

    soma_grower = tested.SomaGrower(tested.Soma((1.0, 2.0, 3.0), 2.0))

    soma_grower._contour_soma = lambda: "contour_function"
    soma_grower._one_point_soma = lambda: "one_point_soma_function"
    soma_grower._original_soma = lambda: "original_soma_function"

    assert_equal(soma_grower.build(method="contour"), "contour_function")
    assert_equal(soma_grower.build(method="one_point"), "one_point_soma_function")
    assert_equal(soma_grower.build(method="original"), "original_soma_function")


def test_grow_soma_types():
    np.random.seed(0)
    g = NeuronGrower(
        {
            "origin": [0, 0, 0],
            "grow_types": [],
            "diameter_params": {"method": "default"},
        },
        {
            "soma": {"size": {"norm": {"mean": 6, "std": 3}}},
            "diameter": {"method": "default"},
        },
    )

    with patch.object(g, "_grow_trunks"):
        # test one soma point
        g._grow_soma(soma_type="one_point")
        assert_array_equal(g.neuron.soma.points, [[0.0, 0.0, 0.0]])
        assert_array_almost_equal(g.neuron.soma.diameters, [22.584314])

        # normal case
        g.soma_grower.soma.points = [[0, 0, 0], [1, 0, 0], [0, 1, 1], [1, 0, 0]]
        g._grow_soma(soma_type="original")
        assert_array_equal(
            g.neuron.soma.points,
            [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, 1.0, 1.0], [1.0, 0.0, 0.0]],
        )
        assert_array_equal(g.neuron.soma.diameters, [0, 0, 0, 0])
        g._grow_soma(soma_type="contour")
        assert_array_almost_equal(
            g.neuron.soma.points,
            [
                [-7.10052, 8.7804, 0.0],
                [-8.961295, -6.870808, 0.0],
                [-6.794974, -9.018932, 0.0],
                [-3.1424513, -10.846097, 0.0],
                [4.315604, -10.434959, 0.0],
                [5.211388, -10.017696, 0.0],
                [1.0, 0.0, 0.0],
            ],
        )


def test_soma_grower():
    # pylint: disable=no-member
    # pylint: disable=unsubscriptable-object
    np.random.seed(0)
    with open(os.path.join(_path, "dummy_distribution.json"), encoding="utf-8") as f:
        distributions = json.load(f)

    with open(os.path.join(_path, "dummy_params.json"), encoding="utf-8") as f:
        params = json.load(f)

    grower = NeuronGrower(input_distributions=distributions, input_parameters=params).grow()

    expected = morphio.Morphology(os.path.join(_path, "dummy_neuron.asc"))

    assert_array_almost_equal(grower.soma.points, expected.soma.points)
    assert_array_almost_equal(grower.root_sections[0].points, expected.root_sections[0].points)
    for sec_actual, sec_expected in zip(grower.iter(), expected.iter()):
        assert_array_almost_equal(sec_actual.points, sec_expected.points)


def test_apical_points():
    # Found apical point
    np.random.seed(0)
    with open(os.path.join(_path, "bio_distribution.json"), encoding="utf-8") as f:
        distributions = json.load(f)

    with open(os.path.join(_path, "bio_path_params.json"), encoding="utf-8") as f:
        params = json.load(f)

    grower = NeuronGrower(input_distributions=distributions, input_parameters=params)
    grower.grow()

    apicals = grower.apical_sections
    expected = [127]
    assert_array_equal(apicals[0], expected)

    # Found apical point
    np.random.seed(0)
    with open(os.path.join(_path, "bio_path_distribution.json"), encoding="utf-8") as f:
        distributions = json.load(f)

    with open(os.path.join(_path, "bio_path_params.json"), encoding="utf-8") as f:
        params = json.load(f)

    grower = NeuronGrower(input_distributions=distributions, input_parameters=params)
    grower.grow()

    apicals = grower.apical_sections
    expected = [60]
    assert_array_equal(apicals, expected)

    # Apical point not found so keep the last bifurcation
    np.random.seed(0)
    with open(os.path.join(_path, "bio_distribution_apical_point.json"), encoding="utf-8") as f:
        distributions = json.load(f)

    with open(os.path.join(_path, "bio_path_params.json"), encoding="utf-8") as f:
        params = json.load(f)

    grower = NeuronGrower(input_distributions=distributions, input_parameters=params)
    grower.grow()

    apicals = grower.apical_sections
    expected = [11]
    assert_array_equal(apicals, expected)


def test_null_orientation():
    np.random.seed(1)
    with open(os.path.join(_path, "dummy_params.json"), encoding="utf-8") as f:
        params = json.load(f)
    params["apical_dendrite"]["orientation"] = [[0, 0, 0], [1, 0, 0]]
    with open(os.path.join(_path, "dummy_distribution.json"), encoding="utf-8") as f:
        distributions = json.load(f)
    N = NeuronGrower(input_distributions=distributions, input_parameters=params)
    with pytest.raises(AssertionError):
        N.grow()



"""Module for handling the calculation of neurite orientations."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import inspect

import numpy as np

from neurots.morphmath import rotation
from neurots.morphmath import sample
from neurots.morphmath.utils import normalize_vectors
from neurots.utils import NeuroTSError

_TWOPI = 2.0 * np.pi


class OrientationManagerBase:
    """Base class that automatically registers orientation modes.

    Args:
        soma (Soma): The soma on which the trees should be attached.
        parameters (dict): The parameters used to compute the orientations.
        distributions (dict): The distributions used to compute the orientations.
        context (any): An object containing contextual information.
        rng (numpy.random.Generator): The random number generator to use.

    .. note::
        To register an orientation mode, derive from this base class and
        create a method with the following signature:

        .. code-block:: python

            def _mode_{name}(self, values_dict, tree_type)
    """

    def __init__(self, soma, parameters, distributions, context, rng):

        self._soma = soma
        self._parameters = parameters
        self._distributions = distributions
        self._context = context
        self._rng = rng

        self._orientations = {}
        self._modes = self._collect_mode_methods()

    def _collect_mode_methods(self):
        """Collects and stores in self._modes the methods, the name of which starts with '_mode_'.

        Returns:
            dict: A dictionary mapping mode names without the '_modes_' prefix to the methods.
        """
        methods = inspect.getmembers(self, predicate=inspect.ismethod)
        return {
            name.replace("_mode_", ""): method
            for name, method in methods
            if name.startswith("_mode_")
        }

    @property
    def mode_names(self):
        """Returns the names of the available modes."""
        return self._modes.keys()

    def get_tree_type_orientations(self, tree_type):
        """Returns the orientations for the specific tree type."""
        return self._orientations.get(tree_type, None)

    def compute_tree_type_orientations(self, tree_type):
        """Computes the orientations for all tree types. It updates the _orientations dictionary.

        Notes:
            Updating a data structure that is accessible by all orientation methods, allows to
            take into account existing orientations for calculating new ones.
        """
        mode_method, values_dict = self._tree_type_method_values(tree_type)
        oris = mode_method(values_dict, tree_type)
        self._orientations[tree_type] = oris
        return oris

    def _tree_type_method_values(self, tree_type):
        """Checks if the orientation entry in parameters has the correct format."""
        params = self._parameters[tree_type]["orientation"]

        if params is None or params.keys() != {"mode", "values"}:
            raise NeuroTSError(
                f"Parameters orientation entry for {tree_type} is not valid"
                f"Expected dict keys {'mode', 'values'}, but got params={params}"
            )

        mode_name = params["mode"]

        if mode_name not in self._modes:
            raise NeuroTSError(
                f"Orientation mode name {mode_name} is not recognized.\n"
                f"Available orientation modes: {self.mode_names}"
            )

        return self._modes[mode_name], params["values"]


class OrientationManager(OrientationManagerBase):
    """Class to generate the tree orientations starting from the soma of the cell.

    Args:
        soma (Soma): The soma on which the trees should be attached.
        parameters (dict): The parameters used to compute the orientations.
        distributions (dict): The distributions used to compute the orientations.
        context (any): An object containing contextual information.
        rng (numpy.random.Generator): The random number generator to use.

    .. note::
        All orientation mode dicts:

        .. code-block:: python

            {
                "mode": "use_predefined",
                "values": {"orientations": [or1, or2, ...]}
            }
            {
                "mode": "sample_pairwise_angles",
                "values": {}
            }
            {
                "mode": "sample_around_primary_orientation",
                "values": {"primary_orientation": [0, 1, 0]}
            }

    """

    def _mode_use_predefined(self, values_dict, tree_type):
        """Returns predefined orientations."""
        assert "orientations" in values_dict, "'orientations' key is missing"
        tree_type_distrs = self._distributions[tree_type]

        # the reason of this sampling is to maintain the pseudorandom
        # sequence of the legacy implementation. Otherwise the functional tests
        # will break because the sequence will be slightly different.
        sample.n_neurites(tree_type_distrs["num_trees"], self._rng)
        return normalize_vectors(np.asarray(values_dict["orientations"], dtype=np.float64))

    def _mode_sample_around_primary_orientation(self, values_dict, tree_type):
        """Sample orientations around a primary direction."""
        tree_type_distrs = self._distributions[tree_type]
        n_orientations = sample.n_neurites(tree_type_distrs["num_trees"], self._rng)

        trunk_absolute_angles = sample.trunk_absolute_angles(
            tree_type_distrs, n_orientations, self._rng
        )
        z_angles = sample.azimuth_angles(tree_type_distrs, n_orientations, self._rng)

        primary_orientation = np.asarray(values_dict["primary_orientation"], dtype=np.float64)
        primary_orientation /= np.linalg.norm(primary_orientation)

        phis, thetas = trunk_absolute_orientation_to_spherical_angles(
            orientation=primary_orientation,
            trunk_absolute_angles=trunk_absolute_angles,
            z_angles=z_angles,
        )

        return spherical_angles_to_orientations(phis, thetas)

    def _mode_sample_pairwise_angles(self, _, tree_type):
        """Returns sampled orientations."""
        tree_type_distrs = self._distributions[tree_type]

        n_orientations = sample.n_neurites(tree_type_distrs["num_trees"], self._rng)

        phi_intervals, interval_n_trees = compute_interval_n_tree(
            self._soma,
            n_orientations,
            self._rng,
        )

        # Create trunks in each interval
        orientations_i = []
        for phi_interval, i_n_trees in zip(phi_intervals, interval_n_trees):
            phis, thetas = trunk_to_spherical_angles(
                sample.trunk_angles(tree_type_distrs, i_n_trees, self._rng),
                sample.azimuth_angles(tree_type_distrs, i_n_trees, self._rng),
                phi_interval,
            )
            orientations_i.append(spherical_angles_to_orientations(phis, thetas))
        return np.vstack(orientations_i)


def spherical_angles_to_orientations(phis, thetas):
    """Compute orientation from spherical angles.

    Args:
        phis (numpy.ndarray): Polar angles.
        thetas (numpy.ndarray): Azimuthal angles.

    Returns:
        numpy.ndarray: The orientation vectors where each row correspnds to a phi-theta pair.
    """
    return np.column_stack(
        (np.cos(phis) * np.sin(thetas), np.sin(phis) * np.sin(thetas), np.cos(thetas))
    )


def points_to_orientations(origin, points):
    """Returns the unit vector that corresponds to the orientation of a point on the soma surface.

    Args:
        origin (numpy.ndarray): The origin of the vectors.
        points (numpy.ndarray): Points to calculate the vectors to.

    Returns:
        numpy.ndarray: Normalized orientations from origin to points.
    """
    return normalize_vectors(points - origin)


def orientations_to_sphere_points(oris, sphere_center, sphere_radius):
    """Compute points on a sphere from the given directions.

    Args:
        oris (numpy.ndarray): Normalized orientation vectors.
        sphere_center (numpy.ndarray): Center of sphere.
        sphere_radius (float): Radius of sphere.

    Returns:
        numpy.ndarray: Points on the surface of the sphere corresponding
        to the given orientations.
    """
    return sphere_center + oris * sphere_radius


def trunk_to_spherical_angles(trunk_angles, z_angles, phi_interval=None):
    """Generate spherical angles from a list of NeuroM angles.

    Args:
        trunk_angles (list[float]): The polar angles (phi in spherical coordinates).
        z_angles (list[float]): The azimuthal angles (theta in spherical coordinates).
        phi_interval (tuple[float, float]): The interval in which the trunks should be added.

    Returns:
        tuple[numpy.ndarray[float], numpy.ndarray[float]]: The phi and theta angles.
    """
    if phi_interval is None:
        phi_interval = (0.0, _TWOPI)

        nb_intervals_min = 0
    else:
        assert len(phi_interval) == 2, "'phi_interval' must be a sequence of 2 elements."
        assert phi_interval[0] < phi_interval[1], "'phi_interval' must be sorted ascending."

        # Add 1 so the equiangle is computed such that angles are not equal to any boundary of the
        # given interval
        nb_intervals_min = 1

    trunk_angles = np.asarray(trunk_angles)
    z_angles = np.asarray(z_angles)

    n_angles = len(trunk_angles)
    sorted_ids = np.argsort(trunk_angles)

    sorted_phi_devs = trunk_angles[sorted_ids]

    thetas = z_angles[sorted_ids]

    equiangle = (phi_interval[1] - phi_interval[0]) / (n_angles + nb_intervals_min)
    phis = np.arange(1, n_angles + 1) * equiangle + sorted_phi_devs + phi_interval[0]

    return phis, thetas


def trunk_absolute_orientation_to_spherical_angles(orientation, trunk_absolute_angles, z_angles):
    """Generate spherical angles from a unit vector and a list of angles.

    Args:
        orientation (list[float]): The orientation vector.
        trunk_absolute_angles (list[float]): The polar angles (phi in spherical coordinates).
        z_angles (list[float]): The azimuthal angles (theta in spherical coordinates).

    Returns:
        tuple[numpy.ndarray[float], numpy.ndarray[float]]: The phi and theta angles.
    """
    # Sort angles
    sort_ids = np.argsort(trunk_absolute_angles)
    sorted_phis = np.asarray(trunk_absolute_angles)[sort_ids]
    sorted_thetas = np.asarray(z_angles)[sort_ids]

    # Convert orientation vector to angles
    phi, theta = rotation.spherical_from_vector(orientation)

    phis = phi + sorted_phis - 0.5 * np.pi
    thetas = theta + sorted_thetas - 0.5 * np.pi

    return phis, thetas


def compute_interval_n_tree(soma, n_trees, rng=np.random):
    """Compute the number of trunks to add between each pair of consecutive existing trunks.

    If points already exist in the soma, the algorithm is the following:

    - build the intervals between each pair of consecutive points.
    - compute the size of each interval.
    - randomly select the interval in which each new point will be added (the intervals are
      weighted by their sizes to ensure the new trunks are created isotropically).
    - count the number of new points in each interval.
    - return the intervals in which at least one point must be added.

    If no point exists in the soma, the interval [0, 2pi] contains all the new trunks.

    Args:
        soma (Soma): The soma on which the trunks should be added.
        n_trees (int): The number of trees that should be added.
        rng (numpy.random.Generator): The random number generator to use.

    Returns:
        tuple[numpy.ndarray[float], numpy.ndarray[int]]: The phi intervals and the number of trees
        in each of them.
    """
    if soma and len(soma.points) > 0:
        # Get angles of existing trunk origins
        phis = []
        for pt in soma.points:
            pt_orientation = soma.orientation_from_point(pt)
            phi, _ = rotation.spherical_from_vector(pt_orientation)
            phis.append(phi)

        phis = sorted(phis)

        # The last interval goes beyond 2 * pi but the function
        # self.soma.add_points_from_trunk_angles can deal with it.
        phis += [phis[0] + _TWOPI]
        phi_intervals = np.column_stack((phis[:-1], phis[1:]))

        # Compute the number of trunks to create in each interval: each interval is weighted by
        # its size to ensure the new trunks are created isotropically.
        sizes = phi_intervals[:, 1] - phi_intervals[:, 0]
        interval_i, interval_i_n_trees = np.unique(
            rng.choice(range(len(sizes)), size=n_trees, p=sizes / sizes.sum()), return_counts=True
        )

        # Keep only intervals with n_trees > 0
        phi_intervals = phi_intervals[interval_i].tolist()
        interval_n_trees = interval_i_n_trees

    else:
        # If there is no existing trunk, we create a None interval
        phi_intervals = [None]
        interval_n_trees = np.array([n_trees])

    return phi_intervals, interval_n_trees



"""NeuroTS class: Grower object that contains the grower functionality."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import copy
import json
import logging

import numpy as np
from morphio.mut import Morphology  # pylint: disable=import-error
from numpy.random import BitGenerator
from numpy.random import Generator
from numpy.random import RandomState
from numpy.random import SeedSequence

from neurots.generate import diametrizer
from neurots.generate import orientations as _oris
from neurots.generate.orientations import OrientationManager
from neurots.generate.soma import Soma
from neurots.generate.soma import SomaGrower
from neurots.generate.tree import TreeGrower
from neurots.morphmath import sample
from neurots.morphmath.utils import normalize_vectors
from neurots.utils import convert_from_legacy_neurite_type
from neurots.validator import validate_neuron_distribs
from neurots.validator import validate_neuron_params

L = logging.getLogger(__name__)

bifurcation_methods = ["symmetric", "bio_oriented", "directional", "bio_smoothed"]


def _load_json(path_or_json):
    """Copy the given data if it is a dictionary or a list or load it if it is a file path."""
    if isinstance(path_or_json, (dict, list)):
        data = copy.deepcopy(path_or_json)
    else:
        with open(path_or_json, encoding="utf-8") as f:
            data = json.load(f)
    return convert_from_legacy_neurite_type(data)


class NeuronGrower:
    """The main class for growing algorithms of neurons.

    A Grower object is a container for a Neuron, encoded in the (groups, points) structure,
    as a morphIO Morphology object. A set of input distributions that store the data
    consumed by the algorithms and the user-selected parameters are also stored.

    Args:
        input_parameters (dict): The user-defined parameters.
        input_distributions (dict): Distributions extracted from biological data.
        context (Any): An object containing contextual information.
        external_diametrizer (Callable): Diametrizer function for external diametrizer module
        skip_validation (bool): If set to ``False``, the parameters and distributions are
            validated.
        rng_or_seed (int or numpy.random.Generator): A random number generator to use. If an
            int is given, it is passed to :func:`numpy.random.default_rng()` to create a new
            random number generator.
        trunk_orientations_class (typing.Generic[OrientationManagerBase]): The class used to
            build the trunk orientation manager. This class should inherit from
            :class:`neurots.generate.orientations.OrientationManagerBase`.
    """

    def __init__(
        self,
        input_parameters,
        input_distributions,
        context=None,
        external_diametrizer=None,
        skip_validation=False,
        rng_or_seed=np.random,
        trunk_orientations_class=OrientationManager,
    ):
        """Constructor of the NeuronGrower class."""
        self.neuron = Morphology()
        self.context = context
        self.skip_validation = skip_validation
        if rng_or_seed is None or isinstance(
            rng_or_seed, (int, np.integer, SeedSequence, BitGenerator)
        ):
            self._rng = np.random.default_rng(rng_or_seed)
        elif isinstance(rng_or_seed, (RandomState, Generator)) or rng_or_seed is np.random:
            self._rng = rng_or_seed
        else:
            raise TypeError(
                "The 'rng_or_seed' argument must be None, np.random or an instance of one of the "
                "following types: [int, SeedSequence, BitGenerator, RandomState, Generator]."
            )

        self.input_parameters = _load_json(input_parameters)
        L.debug("Input Parameters: %s", self.input_parameters)
        self.input_distributions = _load_json(input_distributions)

        # Validate parameters and distributions
        if not skip_validation:
            self.validate_params()
            self.validate_distribs()

        # Consistency check between parameters and distributions
        for tree_type in self.input_parameters["grow_types"]:
            metric1 = self.input_parameters[tree_type].get("metric")
            metric2 = self.input_distributions[tree_type].get("filtration_metric")
            if metric1 not in ["trunk_length", metric2]:
                raise ValueError(
                    "Metric of parameters and distributions is inconsistent:"
                    + f" {metric1} != {metric2}"
                )

        method1 = self.input_parameters["diameter_params"]["method"]
        method2 = self.input_distributions["diameter"]["method"]
        if method1 != method2:
            raise ValueError(
                "Diameters methods of parameters and distributions is inconsistent:"
                + f" {method1} != {method2}"
            )

        if (
            self.input_distributions["diameter"]["method"] == "external"
            and external_diametrizer is None
        ):
            raise ValueError("External diametrizer is missing the diametrizer function.")

        # A list of trees with the corresponding orientations
        # and initial points on the soma surface will be initialized.
        self.active_neurites = []
        self.soma_grower = SomaGrower(
            Soma(
                center=self.input_parameters["origin"],
                radius=sample.soma_size(self.input_distributions, self._rng),
            ),
            context=context,
            rng=self._rng,
        )
        # Create a list to expose apical sections for each apical tree in the neuron,
        # the user can call NeuronGrower.apical_sections to get section IDs whose the last
        # point is the apical point of each generated apical tree.
        self.apical_sections = []

        # initialize diametrizer
        self._init_diametrizer(external_diametrizer=external_diametrizer)

        self._trunk_orientations_class = trunk_orientations_class

    def validate_params(self):
        """Validate the parameter dictionary."""
        validate_neuron_params(self.input_parameters)

    def validate_distribs(self):
        """Validate the distribution dictionary."""
        validate_neuron_distribs(self.input_distributions)

    def next(self):
        """Call the "next" method of each neurite grower."""
        for grower in list(self.active_neurites):
            if grower.end():
                # If tree is an apical, the apical points get appended at the end of growth
                # This will ensure that for each apical tree a relevant apical point,
                # will be exposed to the user as a set of 3D coordinates (x,y,z).
                if (
                    "apical_dendrite" in self.input_parameters["grow_types"]
                    and grower.type == self.input_parameters["apical_dendrite"]["tree_type"]
                ):
                    self.apical_sections.append(grower.growth_algo.apical_section)
                self.active_neurites.remove(grower)
            else:
                grower.next_point()

    def grow(self):
        """Generates a neuron according to the input_parameters and the input_distributions.

        The neuron consists of a soma
        and a list of trees encoded in the h5 format as a set of points
        and groups.

        Returns:
            morphio.mut.Morphology: The grown neuron.
        """
        self._grow_soma()
        while self.active_neurites:
            self.next()  # pylint: disable=E1102
        self._post_grow()
        self._diametrize()
        return self.neuron

    def _post_grow(self):
        """Actions after the morphology has been grown and before its diametrization."""

    def _init_diametrizer(self, external_diametrizer=None):
        """Set a diametrizer function."""
        if self.input_distributions["diameter"]["method"] == "default":
            self._diametrize = lambda: None
            L.warning("No diametrizer provided, so neurons will have default diameters.")
        else:
            if self.input_distributions["diameter"]["method"] == "external":
                if external_diametrizer is None:
                    raise Exception("Please provide an external diametrizer!")
                diam_method = external_diametrizer
            else:
                diam_method = self.input_distributions["diameter"]["method"]

            def _diametrize():
                """Diametrizer function."""
                self.input_distributions["diameter"]["apical_point_sec_ids"] = self.apical_sections
                neurite_types = self.input_parameters.get("diameter_params", {}).get(
                    "neurite_types", None
                )
                if neurite_types is None:
                    neurite_types = self.input_parameters["grow_types"]
                diametrizer.build(
                    self.neuron,
                    self.input_distributions["diameter"],
                    neurite_types=neurite_types,
                    diam_method=diam_method,
                    diam_params=self.input_parameters.get("diameter_params", {}),
                    random_generator=self._rng,
                )

            self._diametrize = _diametrize

    def _convert_orientation2points(self, orientation, n_trees, distr, params):
        """Return soma point from given orientations.

        Checks the type of orientation input and returns the soma points generated by the
        corresponding selection.

        Currently accepted orientations include the following options:
        * list of 3D points: select the orientation externally.
        * ``None``: creates a list of orientations according to the biological distributions.
        * ``from_space``: generates orientations depending on spatial input (not implemented yet).
        """
        # pylint: disable=too-many-locals
        if isinstance(orientation, list):  # Gets major orientations externally
            assert np.all(
                np.linalg.norm(orientation, axis=1) > 0
            ), "Orientations should have non-zero lengths"
            if params.get("trunk_absolute_orientation", False):
                if len(orientation) == 1:

                    orientation = np.asarray(orientation[0], dtype=np.float64)
                    orientation /= np.linalg.norm(orientation)

                    # Pick random absolute angles
                    trunk_absolute_angles = sample.trunk_absolute_angles(distr, n_trees, self._rng)
                    z_angles = sample.azimuth_angles(distr, n_trees, self._rng)

                    phis, thetas = _oris.trunk_absolute_orientation_to_spherical_angles(
                        orientation, trunk_absolute_angles, z_angles
                    )

                    orientations = _oris.spherical_angles_to_orientations(phis, thetas)

                else:
                    raise ValueError("The orientation should contain exactly one point!")
            else:
                if len(orientation) >= n_trees:

                    # TODO: pick orientations randomly?
                    orientations = normalize_vectors(np.asarray(orientation, dtype=np.float64))[
                        :n_trees
                    ]

                else:
                    raise ValueError("Not enough orientation points!")
        elif orientation is None:  # Samples from trunk_angles
            phi_intervals, interval_n_trees = _oris.compute_interval_n_tree(
                self.soma_grower.soma,
                n_trees,
                self._rng,
            )

            # Create trunks in each interval
            orientations_i = []
            for phi_interval, i_n_trees in zip(phi_intervals, interval_n_trees):
                phis, thetas = _oris.trunk_to_spherical_angles(
                    sample.trunk_angles(distr, i_n_trees, self._rng),
                    sample.azimuth_angles(distr, i_n_trees, self._rng),
                    phi_interval,
                )
                orientations_i.append(_oris.spherical_angles_to_orientations(phis, thetas))
            orientations = np.vstack(orientations_i)

        elif orientation == "from_space":
            raise ValueError("Not implemented yet!")
        else:
            raise ValueError("Input orientation format is not correct!")

        pts = self.soma_grower.add_points_from_orientations(orientations)
        return pts

    def _grow_trunks(self):
        """Grow the trunks.

        Generates the initial points of each tree, which depend on the selectedS
        tree types and the soma surface. All the trees start growing from the surface
        of the soma. The outgrowth direction is either specified in the input parameters,
        as ``parameters['type']['orientation']`` or it is randomly chosen according to the
        biological distribution of trunks on the soma surface if ``orientation`` is ``None``.
        """
        tree_types = self.input_parameters["grow_types"]

        legacy_mode = not isinstance(self.input_parameters[tree_types[0]]["orientation"], dict)

        if not legacy_mode:

            trunk_orientations_manager = self._trunk_orientations_class(
                soma=self.soma_grower.soma,
                parameters=self.input_parameters,
                distributions=self.input_distributions,
                context=self.context,
                rng=self._rng,
            )

        for type_of_tree in tree_types:

            params = self.input_parameters[type_of_tree]
            distr = self.input_distributions[type_of_tree]

            if legacy_mode:

                n_trees = sample.n_neurites(distr["num_trees"], random_generator=self._rng)

                if type_of_tree == "basal_dendrite" and n_trees < 2:
                    raise Exception(f"There should be at least 2 basal dendrites (got {n_trees})")

                orientation = params["orientation"]
                points = self._convert_orientation2points(orientation, n_trees, distr, params)

            else:

                orientations = trunk_orientations_manager.compute_tree_type_orientations(
                    type_of_tree
                )
                n_trees = len(orientations)

                if type_of_tree == "basal_dendrite" and n_trees < 2:
                    raise Exception(f"There should be at least 2 basal dendrites (got {n_trees})")

                points = self.soma_grower.add_points_from_orientations(orientations)

            # Iterate over all initial points on the soma and create new trees
            # with a direction and initial_point
            for p in points:
                self.active_neurites.append(
                    TreeGrower(
                        self.neuron,
                        initial_direction=self.soma_grower.soma.orientation_from_point(p),
                        initial_point=p,
                        parameters=params,
                        distributions=distr,
                        skip_validation=self.skip_validation,
                        context=self.context,
                        random_generator=self._rng,
                    )
                )

    def _grow_soma(self, soma_type="contour"):
        """Generates a soma based on the input_distributions.

        The coordinates of the soma contour are retrieved from the trunks.
        """
        self._grow_trunks()

        points, diameters = self.soma_grower.build(soma_type)
        self.neuron.soma.points = points
        self.neuron.soma.diameters = diameters



"""Test neurots.validator code."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=missing-function-docstring
# pylint: disable=redefined-outer-name
import json
from pathlib import Path

import pytest

import neurots.validator as tested

DATA = Path(__file__).parent.resolve() / "data"


@pytest.fixture
def dummy_params():
    with (DATA / "params2.json").open(encoding="utf-8") as f:
        data = json.load(f)
    return data


@pytest.fixture
def dummy_distribs():
    with (DATA / "dummy_distribution.json").open(encoding="utf-8") as f:
        data = json.load(f)
    return data


@pytest.fixture
def interneuron_distribs():
    with (DATA / "dummy_interneuron_distribution.json").open(encoding="utf-8") as f:
        data = json.load(f)
    return data


def test_validate_params(dummy_params):
    tested.validate_neuron_params(dummy_params)
    dummy_params["apical_dendrite"]["orientation"] = None
    tested.validate_neuron_params(dummy_params)


class TestValidateParams:
    """Test the parameter validation."""

    def test_default(self, dummy_params):
        tested.validate_neuron_params(dummy_params)

    def test_none_orientation(self, dummy_params):
        dummy_params["apical_dendrite"]["orientation"] = None
        tested.validate_neuron_params(dummy_params)

    def test_different_external_method(self, dummy_params):
        # If method != external it should not have any other entry than 'method'
        dummy_params["diameter_params"] = {"method": "default"}
        tested.validate_neuron_params(dummy_params)

    def test_default_diameter_unknown_key(self, dummy_params):
        dummy_params["diameter_params"] = {
            "method": "default",
            "other key": "any value",
        }
        with pytest.raises(tested.ValidationError):
            tested.validate_neuron_params(dummy_params)

    def test_M1_diameter_unknown_key(self, dummy_params):
        dummy_params["diameter_params"] = {"method": "M1", "other key": "any value"}
        with pytest.raises(tested.ValidationError):
            tested.validate_neuron_params(dummy_params)

    def test_external_diameter(self, dummy_params):
        # If method == external it may have any other entry than 'method'
        dummy_params["diameter_params"] = {"method": "external"}
        tested.validate_neuron_params(dummy_params)

    def test_external_diameter_unknown_key(self, dummy_params):
        dummy_params["diameter_params"] = {
            "method": "external",
            "other key": "any value",
        }
        tested.validate_neuron_params(dummy_params)

    def test_orientation(self, dummy_params):
        # It must be a list of vectors, not a single one
        dummy_params["apical_dendrite"]["orientation"] = [0, 0, 0]
        with pytest.raises(tested.ValidationError):
            tested.validate_neuron_params(dummy_params)

    def test_unknown_root_key(self, dummy_params):
        # Unknown parameters can be added at the root
        dummy_params["unknown_param"] = 0
        tested.validate_neuron_params(dummy_params)


def test_empty_params():
    data = {
        "apical_dendrite": {},
        "axon": {},
        "basal_dendrite": {},
        "diameter_params": {"method": "M5"},
        "grow_types": [],
        "origin": [0.0, 0.0, 0.0],
    }
    tested.validate_neuron_params(data)


def test_validate_neuron_distribs(dummy_distribs, interneuron_distribs):
    tested.validate_neuron_distribs(dummy_distribs)
    tested.validate_neuron_distribs(interneuron_distribs)



"""Test neurots.astrocyte.context.EndfeetTargets code."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=missing-function-docstring
import numpy as np
from numpy import testing as npt

from neurots.astrocyte import context as tested

TARGET_POINTS = np.array([[0.0, 1.0, 2.0], [2.0, 3.0, 4.0], [5.0, 6.0, 7.0]])


def test_constructor():
    targets = tested.EndfeetTargets(TARGET_POINTS)
    npt.assert_allclose(targets.points, TARGET_POINTS)
    npt.assert_array_equal(targets.active, True)
    npt.assert_equal(len(targets), 3)


def test_active_points():
    targets = tested.EndfeetTargets(TARGET_POINTS)

    targets.active[1] = False
    npt.assert_allclose(targets.active_points, TARGET_POINTS[(0, 2), :])

    targets.active[:] = False
    npt.assert_equal(len(targets.active_points), 0)



"""Input distributions."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging

import tmd
from neurom import NeuriteType
from neurom import load_morphologies

from neurots.extract_input import from_diameter
from neurots.extract_input.from_neurom import number_neurites
from neurots.extract_input.from_neurom import soma_data
from neurots.extract_input.from_neurom import trunk_neurite
from neurots.extract_input.from_TMD import persistent_homology_angles
from neurots.utils import format_values
from neurots.utils import neurite_type_warning

L = logging.getLogger(__name__)


def _append_dicts(*args):
    """Merge all dicts into the first one."""
    ret = args[0]
    for other_dict in args[1:]:
        ret.update(other_dict)
    return ret


def distributions(
    filepath,
    neurite_types=None,
    threshold_sec=2,
    diameter_input_morph=None,
    feature="path_distances",
    diameter_model=None,
):
    """Extracts the input distributions from an input population.

    The population is defined by a directory of swc or h5 files.

    Args:
        filepath (str): the morphology file.
        neurite_types (list[str]): the neurite types to consider.
        threshold_sec (int): defines the minimum accepted number of terminations.
        diameter_input_morph (str): if input set of morphologies is provided it will be used for the
            generation of diameter model, if no input is provided no diameter model will be
            generated.
        feature (str): defines the TMD feature that will be used to extract the persistence barcode
            (can be `radial_distances`, `path_distances` or `trunk_length`). It is also possible to
            define one different feature per neurite type using a dict like
            ``{<neurite type 1>: <feature 1>, ...}``.
        diameter_model (str): model for diameters, internal models are `M1`, `M2`, `M3`, `M4` and
            `M5`. Can be set to `external` for external model.

    Returns:
        dict: The input distributions.
    """
    if neurite_types is None:
        neurite_types = ["basal_dendrite", "apical_dendrite", "axon"]

    for i, neurite_type in enumerate(neurite_types):
        if neurite_type in ("basal", "apical"):
            neurite_type_warning(neurite_type)
            neurite_types[i] = neurite_type + "_dendrite"

    pop_tmd = tmd.io.load_population(filepath, use_morphio=True)
    pop_nm = load_morphologies(filepath)

    input_distributions = {"soma": {}, "basal_dendrite": {}, "apical_dendrite": {}, "axon": {}}
    input_distributions["soma"] = soma_data(pop_nm)

    if diameter_input_morph is None:
        diameter_input_morph = filepath

    if isinstance(diameter_model, str):
        input_distributions["diameter"] = from_diameter.model(
            load_morphologies(diameter_input_morph)
        )
        input_distributions["diameter"]["method"] = diameter_model

    elif hasattr(diameter_model, "__call__"):
        input_distributions["diameter"] = diameter_model(load_morphologies(diameter_input_morph))
        input_distributions["diameter"]["method"] = "external"
    else:
        input_distributions["diameter"] = {}
        input_distributions["diameter"]["method"] = "default"
        L.warning("No valid diameter model provided, so we will not generate a distribution.")

    for neurite_type in neurite_types:
        if isinstance(feature, str):
            type_feature = feature
        else:
            type_feature = feature.get(neurite_type, "path_distances")
        nm_type = getattr(NeuriteType, neurite_type)
        input_distributions[neurite_type] = _append_dicts(
            trunk_neurite(pop_nm, nm_type),
            number_neurites(pop_nm, nm_type),
        )
        if type_feature in ["path_distances", "radial_distances"]:
            _append_dicts(
                input_distributions[neurite_type],
                persistent_homology_angles(
                    pop_tmd,
                    threshold=threshold_sec,
                    neurite_type=neurite_type,
                    feature=type_feature,
                ),
                {"filtration_metric": type_feature},
            )
    return format_values(input_distributions)



"""A JSON schema validator for parameters and distributions.

Visit https://json-schema.org/understanding-json-schema/ for more information about JSON schemas.
"""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json
from pathlib import Path

import jsonschema
import pkg_resources

SCHEMA_PATH = pkg_resources.resource_filename("neurots", "schemas")

with Path(SCHEMA_PATH, "parameters.json").open(encoding="utf-8") as f:
    PARAMS_SCHEMA = json.load(f)

with Path(SCHEMA_PATH, "distributions.json").open(encoding="utf-8") as f:
    DISTRIBS_SCHEMA = json.load(f)


class ValidationError(Exception):
    """Exception raised when a JSON object is not valid according to a given schema."""


def _format_error(error):
    return f"""In [{"->".join(error.absolute_path)}]: {error.message}"""


def validate(instance, schema):
    """Validate a JSON object according to a given schema."""
    validator = jsonschema.Draft7Validator(schema)
    errors = sorted(validator.iter_errors(instance), key=lambda e: e.path)
    messages = []
    for error in errors:
        if error.context:
            for suberror in error.context:
                messages.append(_format_error(suberror))
        else:
            messages.append(_format_error(error))

    if messages:
        raise ValidationError("\n".join(messages))


def validate_neuron_params(data):
    """Validate parameter dictionary."""
    validate(data, PARAMS_SCHEMA)


def validate_neuron_distribs(data):
    """Validate distribution dictionary."""
    validate(data, DISTRIBS_SCHEMA)



"""Point cloud class for synapses."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging

import numpy as np
from scipy.spatial import KDTree

from neurots.morphmath.utils import norm as vectorial_norm

L = logging.getLogger(__name__)


MIN_EPS = -1e-6

# pylint: disable=unsubscriptable-object


class PointCloud:
    """Point cloud data structure with internal gridding for ball and nearest neighbor queries.

    The points that are removed are not removed from memory, but invalidated.

    Args:
        point_array (np.ndarray):
            Array of 3D points to store in the point cloud.
        cutoff_radius (float):
            The radius from which the bin size of the grid will be calculated.
        removal_radius (float):
            Radius that will be used for default removal. See attribute below.

    Attributes:
        default_removal_radius (float):
            A default radius that will be used for the remove_points_around if
            an explicit value is not passed in the method.
        default_radius_of_influence (float):
            A default radius for the calculation of the average direction if
            an explicit value is not passed in the method.
    """

    def __init__(self, points):
        self._tree = KDTree(np.asarray(points, dtype=np.float32), copy_data=False)
        self._available = np.ones(len(points), dtype=bool)

    @property
    def points(self):
        """Returns point cloud points."""
        return self._tree.data

    @property
    def available_ids(self):
        """Returns ids of available points."""
        return np.where(self._available)[0]

    @property
    def removed_ids(self):
        """Returns ids of removed points."""
        return np.where(~self._available)[0]

    @property
    def removed_points(self):
        """Returns removed points."""
        return self.points[self.removed_ids]

    @property
    def available_points(self):
        """Returns available points."""
        return self.points[self.available_ids]

    def ball_query(self, point, radius):
        """Ball query around point with radius."""
        indices = np.fromiter(
            self._tree.query_ball_point(point, radius, return_sorted=False, return_length=False),
            dtype=np.int64,
        )
        return indices[self._available[indices]]

    def partial_ball_query(self, point, radius, direction, cap_angle_front, cap_angle_back):
        """Truncated ball query."""
        ids = self.ball_query(point, radius)

        if ids.size == 0:
            return ids

        dots = (self.points[ids] - point).dot(direction)

        dot_above = radius * np.cos(cap_angle_front)
        dot_below = -radius * np.cos(cap_angle_back)

        return ids[(dots > dot_below) & (dots < dot_above)]

    def upper_half_ball_query(self, point, radius, direction):
        """Hemisphere query around point directed to direction."""
        return self.partial_ball_query(point, radius, direction, 0.0, 0.5 * np.pi)

    def nearest_neighbor(self, point, radius):
        """Get the nearest neighbor to the point with cuttoff radius."""
        _, index = self._tree.query(point, k=1, distance_upper_bound=radius)
        if index < self._available.size and self._available[index]:
            return index
        return None

    def nearest_neighbor_direction(self, point, radius):
        """Get nearest neighbor direction."""
        point_id = self.nearest_neighbor(point, radius)

        if point_id is None:
            return None

        vector = self.points[point_id] - point
        return vector / vectorial_norm(vector)

    def remove_ids(self, point_ids):
        """Remove points ids."""
        self._available[point_ids] = False

    def remove_points_around(self, point, radius):
        """Remove the points in the sphere located at point with removal_radius."""
        point_ids = self.ball_query(point, radius)
        self.remove_ids(point_ids)
        return point_ids

    def remove_hemisphere(self, point, direction, radius):
        """Remove hemisphere that points to direction."""
        point_ids = self.ball_query(point, radius)

        vectors = self.points[point_ids] - point
        point_ids = point_ids[vectors.dot(direction) > MIN_EPS]

        self.remove_ids(point_ids)
        return point_ids



"""Extracts the distributions associated with NeuroM module."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import neurom as nm
import numpy as np
from neurom import stats


def transform_distr(opt_distr):
    """Transform distributions.

    Args:
        opt_distr (neurom.stats.FitResults): The fitted distribution.

    Returns:
        dict: A dictionary whose structure depends on the type of distribution:

        * if `type == "norm"`:

        .. code-block:: bash

            {
                "norm": {
                    "mean": <mean value>,
                    "std": <std value>
                }
            }

        * if `type == "uniform"`:

        .. code-block:: bash

            {
                "uniform": {
                    "min": <min value>,
                    "max": <max value>
                }
            }

        * if `type == "expon"`:

        .. code-block:: bash

            {
                "expon": {
                    "loc": <loc value>,
                    "lambda": <lambda value>
                }
            }
    """
    if opt_distr.type == "norm":
        return {"norm": {"mean": opt_distr.params[0], "std": opt_distr.params[1]}}
    elif opt_distr.type == "uniform":
        return {
            "uniform": {
                "min": opt_distr.params[0],
                "max": opt_distr.params[1] + opt_distr.params[0],
            }
        }
    elif opt_distr.type == "expon":
        return {"expon": {"loc": opt_distr.params[0], "lambda": 1.0 / opt_distr.params[1]}}
    return None


def soma_data(pop):
    """Extract soma size.

    Args:
        pop (neurom.core.population.Population): The given population.

    Returns:
        dict: A dictionary with the following structure:

        .. code-block:: bash

            {
                "size": <the soma size>
            }
    """
    # Extract soma size as a normal distribution
    # Returns a dictionary with the soma information
    soma_size = nm.get("soma_radius", pop)
    ss = stats.fit(soma_size, distribution="norm")

    return {"size": transform_distr(ss)}


def trunk_neurite(pop, neurite_type=nm.BASAL_DENDRITE, bins=30):
    """Extract the trunk data for a specific tree type.

    Args:
        pop (neurom.core.population.Population): The given population.
        neurite_type (neurom.core.types.NeuriteType): Consider only the neurites of this type.
        bins (int or list[int] or str, optional): The bins to use (this pararmeter is passed to
            :func:`numpy.histogram`).

    Returns:
        dict: A dictionary with the following structure:

        .. code-block:: bash

            {
                "trunk": {
                    "orientation_deviation": {
                        "data": {
                            "bins": <bin values>,
                            "weights": <weights>
                        }
                    },
                    "azimuth": {
                        "inuform": {
                            "min": <min value>,
                            "max": <max value>
                        }
                    },
                    "absolute_elevation_deviation": {
                        "data": {
                            "bins": <bin values>,
                            "weights": <weights>
                        }
                    }
                }
            }
    """
    angles = [nm.get("trunk_angles", neuron, neurite_type=neurite_type) for neuron in pop]
    angles = np.concatenate(angles, axis=0)
    angle_heights, angle_bins = np.histogram(angles, bins=bins)

    # Extract trunk relative orientations to resample
    actual_angle_bins = (angle_bins[1:] + angle_bins[:-1]) / 2.0

    elevations = [
        nm.get("trunk_origin_elevations", neuron, neurite_type=neurite_type) for neuron in pop
    ]
    elevations = np.concatenate(elevations, axis=0)
    elevation_heights, elevation_bins = np.histogram(elevations, bins=bins)

    # Extract trunk absolute orientations to resample
    actual_elevation_bins = (elevation_bins[1:] + elevation_bins[:-1]) / 2.0

    return {
        "trunk": {
            "orientation_deviation": {
                "data": {"bins": actual_angle_bins.tolist(), "weights": angle_heights.tolist()}
            },
            "azimuth": {"uniform": {"min": np.pi, "max": 0.0}},
            "absolute_elevation_deviation": {
                "data": {
                    "bins": actual_elevation_bins.tolist(),
                    "weights": elevation_heights.tolist(),
                }
            },
        }
    }


def number_neurites(pop, neurite_type=nm.BASAL_DENDRITE):
    """Extract the number of trees for a specific tree type from a given population.

    Args:
        pop (neurom.core.population.Population): The given population.
        neurite_type (neurom.core.types.NeuriteType): Consider only the neurites of this type.

    Returns:
        dict: A dictionary with the following structure:

        .. code-block:: bash

            {
                "num_trees": {
                    "data": {
                        "bins": <bin values>,
                        "weights": <weights>
                    }
                }
            }
    """
    # Extract number of neurites as a precise distribution
    # The output is given in integer numbers which are
    # the permitted values for the number of trees.
    nneurites = np.asarray(
        nm.get("number_of_neurites", pop, neurite_type=neurite_type), dtype=np.int32
    )
    # Clean the data from single basal trees cells
    if neurite_type == nm.BASAL_DENDRITE and len(np.where(nneurites == 1)[0]) > 0:
        nneurites[np.where(nneurites == 1)[0]] = 2
        print(
            "Warning, input population includes cells with single basal trees! "
            + "The distribution has been altered to include 2 basals minimum."
        )

    heights, bins = np.histogram(
        nneurites, bins=np.arange(np.min(nneurites), np.max(nneurites) + 2)
    )

    # pylint: disable=no-member
    return {"num_trees": {"data": {"bins": bins[:-1].tolist(), "weights": heights.tolist()}}}



"""Module to extract morphometrics about diameters of cells."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections import defaultdict
from itertools import chain

import numpy as np
from neurom import get
from neurom.core.morphology import Section
from neurom.core.morphology import iter_neurites
from neurom.morphmath import segment_length
from neurom.morphmath import segment_radius

from neurots.utils import NeuroTSError

default_model = {"Rall_ratio": 3.0 / 2.0, "siblings_ratio": 1.0}


def _check(data):
    """Check if data in dictionary are empty."""
    for key, val in data.items():
        if len(val) == 0:
            raise NeuroTSError(f"Empty distribution for diameter key: {key}")


def section_mean_taper(s):
    """Computes the mean tapering of a section."""
    min_diam = min(s.points[:, 3])

    di_li = sum(
        segment_radius([s.points[i], s.points[i + 1]])
        * segment_length([s.points[i], s.points[i + 1]])
        for i in range(len(s.points) - 1)
    )

    return (di_li - min_diam * s.length) / s.length


def terminal_diam(tree):
    """Returns the model for the terminations."""
    mean_diam = np.mean(tree.points[:, 3])
    term_diam = [
        2.0 * t.points[-1, 3]
        for t in Section.ileaf(next(tree.iter_sections()))
        if t.points[-1, 3] < 1.2 * mean_diam
    ]

    return term_diam


def section_taper(tree):
    """Returns the tapering of the *diameters* within the sections of a tree."""
    # Exclude the trunk = first section, taper should not be x2 because it is relative
    tapers = [section_mean_taper(s) for s in tree.iter_sections()]
    return [taper for taper in tapers if taper][1:]


def section_trunk_taper(tree):
    """Returns the tapering of the *diameters* within the sections of a tree."""
    # Exclude the trunk = first section, taper should not be x2 because it is relative
    return section_mean_taper(next(tree.iter_sections()))


def model(input_object):
    """Measure the statistical properties of input_object's diameters and outputs a diameter_model.

    Input can be a population of neurons, or a single neuron.
    """
    values = {}

    tapers = defaultdict(list)
    trunk_tapers = defaultdict(list)
    term_diams = defaultdict(list)
    trunk_diams = defaultdict(list)

    for neurite in iter_neurites(input_object):
        tapers[neurite.type].append(section_taper(neurite))
        trunk_tapers[neurite.type].append(section_trunk_taper(neurite))
        term_diams[neurite.type].append(terminal_diam(neurite))
        trunk_diams[neurite.type].append(2.0 * np.max(get("segment_radii", neurite)))

    for neurite_type in tapers:
        key = neurite_type.name

        taper_c = np.array(list(chain(*tapers[neurite_type])))
        trunk_taper = np.array(trunk_tapers[neurite_type])

        # Keep only positive, non-zero taper rates
        taper_c = taper_c[np.where(taper_c > 0.00001)[0]]
        trunk_taper = trunk_taper[np.where(trunk_taper >= 0.0)[0]]

        values[key] = {
            "taper": taper_c.tolist(),
            "term": list(chain(*term_diams[neurite_type])),
            "trunk": trunk_diams[neurite_type],
            "trunk_taper": trunk_taper.tolist(),
        }

        _check(values[key])
        values[key].update(default_model)

    return values



"""Test neurots.generate.orientations code."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=missing-function-docstring
# pylint: disable=protected-access
import inspect

import numpy as np
import pytest
from numpy import testing as npt

from neurots.generate import orientations as tested
from neurots.generate.soma import Soma
from neurots.utils import NeuroTSError


def test_orientation_manager__constructor():

    soma = "soma"
    parameters = "parameters"
    distributions = "distributions"
    context = "context"
    rng = "rng"

    om = tested.OrientationManager(
        soma=soma,
        parameters=parameters,
        distributions=distributions,
        context=context,
        rng=rng,
    )

    assert om._parameters is parameters
    assert om._distributions is distributions
    assert om._context is context
    assert om._rng is rng

    assert len(om._orientations) == 0

    # check if all methods are bound
    methods = inspect.getmembers(om)
    mode_methods = [(name, method) for name, method in methods if name.startswith("_mode_")]
    expected_modes = {name.replace("_mode_", ""): method for name, method in mode_methods}

    # check that the correct names have been bound to the
    # respective method names
    for mode, method in om._modes.items():
        assert mode == method.__name__.replace(
            "_mode_", ""
        ), f"Mode name mismatch: {mode} -> {method.__name__}"

    # then check that all there are no new methods that are unregistered
    str_modes = "\n".join(om._modes)
    str_expected_modes = "\n".join(expected_modes)

    assert om._modes == expected_modes, (
        "Not all modes are bound.\n"
        f"\nActual   :\n{str_modes}\n"
        f"\nExpected :\n{str_expected_modes}\n"
    )

    assert set(om.mode_names) == set(expected_modes.keys())


def test_orientation_manager__mode_use_predefined():

    parameters = {
        "grow_types": ["john"],
        "john": {
            "orientation": {
                "mode": "use_predefined",
                "values": {"orientations": [[0.0, 1.0, 0.0], [1.0, 0.0, 0.0]]},
            }
        },
    }

    distributions = {
        "john": {
            "num_trees": {
                "data": {"bins": [2], "weights": [1]},
            }
        }
    }

    om = tested.OrientationManager(
        soma=None,
        parameters=parameters,
        distributions=distributions,
        context=None,
        rng=np.random.default_rng(seed=0),
    )

    for tree_type in parameters["grow_types"]:
        om.compute_tree_type_orientations(tree_type)

    actual = om.get_tree_type_orientations("john")
    expected = np.array([[0.0, 1.0, 0.0], [1.0, 0.0, 0.0]])

    npt.assert_allclose(actual, expected)


def test_orientation_manager__tree_type_method_values():

    parameters = {
        "grow_types": ["john"],
        "john": {
            "orientation": {
                "mode": "use_predefined",
                "values": {"orientations": [[0.0, 1.0, 0.0], [1.0, 0.0, 0.0]]},
            }
        },
    }

    om = tested.OrientationManager(
        soma=None, parameters=parameters, distributions=None, context=None, rng=None
    )

    method, values = om._tree_type_method_values("john")

    assert method == om._mode_use_predefined  # pylint: disable=comparison-with-callable
    assert values == {"orientations": [[0.0, 1.0, 0.0], [1.0, 0.0, 0.0]]}

    # check that method exists
    parameters["john"]["orientation"]["mode"] = "non_existent_method"
    with pytest.raises(NeuroTSError):
        om.compute_tree_type_orientations("john")

    # check that config keys are correct
    parameters["john"]["orientation"]["random_key"] = "lol"
    with pytest.raises(NeuroTSError):
        om.compute_tree_type_orientations("john")

    # check that config is not empty
    parameters["john"]["orientation"] = {}
    with pytest.raises(NeuroTSError):
        om.compute_tree_type_orientations("john")

    # check that orientation is not None
    parameters["john"]["orientation"] = None
    with pytest.raises(NeuroTSError):
        om.compute_tree_type_orientations("john")


def test_orientation_manager__mode_sample_around_primary_orientation():

    parameters = {
        "grow_types": ["john"],
        "john": {
            "orientation": {
                "mode": "sample_around_primary_orientation",
                "values": {"primary_orientation": [0.0, 0.0, 1.0]},
            }
        },
    }

    distributions = {
        "john": {
            "num_trees": {
                "data": {"bins": [2], "weights": [1]},
            },
            "trunk": {
                "azimuth": {"data": {"bins": [np.pi], "weights": [1]}},
                "absolute_elevation_deviation": {"data": {"bins": [np.pi * 0.5], "weights": [1]}},
            },
        }
    }

    om = tested.OrientationManager(
        soma=None,
        parameters=parameters,
        distributions=distributions,
        context=None,
        rng=np.random.default_rng(seed=0),
    )

    for tree_type in parameters["grow_types"]:
        om.compute_tree_type_orientations(tree_type)

    actual = om.get_tree_type_orientations("john")
    expected = np.array([[1.0, 0.0, 0.0], [1.0, 0.0, 0.0]])

    npt.assert_allclose(actual, expected, atol=1e-6)


def test_orientation_manager__mode_sample_pairwise_angles():

    parameters = {
        "grow_types": ["john"],
        "john": {"orientation": {"mode": "sample_pairwise_angles", "values": {}}},
    }

    distributions = {
        "john": {
            "num_trees": {
                "data": {"bins": [3], "weights": [1]},
            },
            "trunk": {
                "azimuth": {"data": {"bins": [np.pi / 2], "weights": [1]}},
                "orientation_deviation": {"data": {"bins": [0], "weights": [1]}},
            },
        }
    }

    om = tested.OrientationManager(
        soma=None,
        parameters=parameters,
        distributions=distributions,
        context=None,
        rng=np.random.default_rng(seed=0),
    )

    for tree_type in parameters["grow_types"]:
        om.compute_tree_type_orientations(tree_type)

    actual = om.get_tree_type_orientations("john")
    expected = np.array([[-0.5, np.sin(np.pi / 3), 0], [-0.5, -np.sin(np.pi / 3), 0], [1, 0, 0]])

    npt.assert_allclose(actual, expected, atol=1e-6)

    # Test with existing neurites in soma
    soma = Soma(
        (0.0, 0.0, 0.0),
        6.0,
        [
            [-6.0, 0.0, 0.0],
            [6.0, 0.0, 0.0],
            [0.0, 6.0, 0.0],
            [0.0, -6.0, 0.0],
        ],
    )

    om = tested.OrientationManager(
        soma=soma,
        parameters=parameters,
        distributions=distributions,
        context=None,
        rng=np.random.default_rng(seed=0),
    )

    for tree_type in parameters["grow_types"]:
        om.compute_tree_type_orientations(tree_type)

    actual = om.get_tree_type_orientations("john")
    expected = np.array(
        [
            [0.5, -np.sin(2 * np.pi / 6), 0],
            [np.sin(2 * np.pi / 6), -0.5, 0],
            [np.sin(np.pi / 4), np.sin(np.pi / 4), 0],
        ]
    )

    npt.assert_allclose(actual, expected, atol=1e-6)


def test_spherical_angles_to_orientations():

    phis = [0.5 * np.pi, np.pi, np.pi]

    thetas = [0.5 * np.pi, np.pi, 0.5 * np.pi]

    expected_orientations = [[0.0, 1.0, 0.0], [0.0, 0.0, -1.0], [-1.0, 0.0, 0.0]]

    npt.assert_allclose(
        tested.spherical_angles_to_orientations(phis, thetas),
        expected_orientations,
        atol=1e-6,
    )


def test_points_to_orientations():

    origin = np.array([0.0, 0.0, 0.0])
    points = np.array([[2.0, 0.0, 0.0], [0.0, 0.0, 3.0]])

    actual = tested.points_to_orientations(origin, points)
    npt.assert_allclose(actual, [[1.0, 0.0, 0.0], [0.0, 0.0, 1.0]])


def test_orientations_to_sphere_points():

    sphere_center = np.array([0.0, 0.0, 1.0])
    sphere_radius = 0.2

    oris = np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])

    expected_points = np.array([[0.2, 0.0, 1.0], [0.0, 0.2, 1.0], [0.0, 0.0, 1.2]])

    npt.assert_allclose(
        tested.orientations_to_sphere_points(oris, sphere_center, sphere_radius),
        expected_points,
    )



"""Test neurots.morphmath.sample code."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=missing-function-docstring
# pylint: disable=protected-access
import numpy as np
import pytest
from numpy.testing import assert_equal

from neurots.morphmath import sample


def test_Distr():
    # pylint: disable=too-many-statements
    # Test distributions
    assert_equal(sample.Distr.norm({"mean": 1, "std": 0.5}), ("standard_normal", 1, 0.5))
    assert_equal(sample.Distr.uniform({"min": 1, "max": 1.25}), ("uniform", 1, 0.25))
    assert_equal(sample.Distr.expon({"loc": 1, "lambda": 2}), ("standard_exponential", 1, 0.5))

    # Setup normal distribution
    np.random.seed(0)
    mt = np.random.MT19937()
    mt._legacy_seeding(0)  # Use legacy seeding to get the same result as with np.random.seed(0)
    rng = np.random.RandomState(mt)
    mt2 = np.random.MT19937()
    mt2._legacy_seeding(0)
    rng_new = np.random.Generator(mt2)

    params = {"norm": {"mean": 9.024144162609812, "std": 3.5462697985669935}}
    soma_d = sample.Distr(params)
    soma_d_rng = sample.Distr(params, random_generator=rng)
    soma_d_new = sample.Distr(params, random_generator=rng_new)

    # Test draw_positive
    val = soma_d.draw_positive()
    val_rng = soma_d_rng.draw_positive()
    val_new = soma_d_new.draw_positive()
    assert_equal(val, 15.279949720206192)
    assert_equal(val_rng, val)
    assert_equal(val_new, 10.356455608693715)

    assert_equal(soma_d.draw_positive(), 10.443209585321375)
    assert_equal(soma_d_rng.draw_positive(), 10.443209585321375)
    assert_equal(soma_d_new.draw_positive(), 14.207311245358344)

    # Test draw
    assert_equal(soma_d.draw(), 12.495013116354336)
    assert_equal(soma_d_rng.draw(), 12.495013116354336)
    assert_equal(soma_d_new.draw(), 18.201541834082416)

    # Setup uniform distribution
    params = {"uniform": {"min": 20, "max": 30}}
    soma_d = sample.Distr(params)
    soma_d_rng = sample.Distr(params, random_generator=rng)
    soma_d_new = sample.Distr(params, random_generator=rng_new)

    # Test draw_positive
    val = soma_d.draw_positive()
    val_rng = soma_d_rng.draw_positive()
    soma_d_new.draw_positive()  # This call is just to align random states of rng and rng_new
    val_new = soma_d_new.draw_positive()
    assert_equal(val, 24.236547993389046)
    assert_equal(val_rng, val)
    assert_equal(val_new, val)

    expected = 26.45894113066656
    assert_equal(soma_d.draw_positive(), expected)
    assert_equal(soma_d_rng.draw_positive(), expected)
    assert_equal(soma_d_new.draw_positive(), expected)

    # Test draw
    expected = 24.375872112626926
    assert_equal(soma_d.draw(), expected)
    assert_equal(soma_d_rng.draw(), expected)
    assert_equal(soma_d_new.draw(), expected)

    # Setup exponential distribution
    params = {"expon": {"loc": 10, "lambda": 5}}
    soma_d = sample.Distr(params)
    soma_d_rng = sample.Distr(params, random_generator=rng)
    soma_d_new = sample.Distr(params, random_generator=rng_new)

    # Test draw_positive
    val = soma_d.draw_positive()
    val_rng = soma_d_rng.draw_positive()
    val_new = soma_d_new.draw_positive()
    assert_equal(val, 10.444704882606532)
    assert_equal(val_rng, val)
    assert_equal(val_new, 10.461004518163628)

    assert_equal(soma_d.draw_positive(), 10.662982436410763)
    assert_equal(soma_d_rng.draw_positive(), 10.662982436410763)
    assert_equal(soma_d_new.draw_positive(), 10.46145745042702)

    # Test draw
    assert_equal(soma_d.draw(), 10.09672042018045)
    assert_equal(soma_d_rng.draw(), 10.09672042018045)
    assert_equal(soma_d_new.draw(), 10.092614156264565)

    # Setup data distribution
    params = {"data": {"weights": [0.1, 0.9], "bins": [1, 2]}}
    soma_d = sample.Distr(params)
    assert_equal(soma_d.draw_positive(), 2)

    soma_d_rng = sample.Distr(params, random_generator=rng)
    assert_equal(soma_d_rng.draw_positive(), 2)

    # Setup negative loc distribution
    params = {"uniform": {"min": -10, "max": -10}}
    soma_d = sample.Distr(params)
    with pytest.raises(ValueError):
        soma_d.draw_positive()

    # Setup negative val distribution
    params = {"uniform": {"min": -50, "max": 10}}
    soma_d = sample.Distr(params, random_generator=rng)
    assert_equal(soma_d.draw_positive(), 5.535798297559666)


def test_soma_size():
    np.random.seed(0)
    rng = np.random.default_rng(0)
    params = {"soma": {"size": {"norm": {"mean": 9.024144162609812, "std": 3.5462697985669935}}}}

    val1 = sample.soma_size(params)
    assert_equal(val1, 15.279949720206192)

    val1_rng = sample.soma_size(params, random_generator=rng)
    assert_equal(val1_rng, 9.470017448440464)



"""Test neurots.morphmath.point_array code."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=missing-function-docstring
# pylint: disable=redefined-outer-name
# pylint: disable=protected-access
import numpy as np
import pytest
from numpy import testing as npt

from neurots.morphmath import point_array as _pa


@pytest.fixture
def dynamic_array():

    initial_capacity = 3
    resize_factor = 2

    array = _pa.DynamicPointArray(initial_capacity, resize_factor)

    return array


def test_dynamic_point_array_contructor(dynamic_array):
    assert len(dynamic_array) == 0

    npt.assert_allclose(dynamic_array.data, np.empty(shape=(0, 3)))

    assert dynamic_array._size == 0
    assert dynamic_array._capacity == 3


def test_dynamic_point_array_append(dynamic_array):
    p0 = np.random.random(3)
    dynamic_array.append(p0)
    npt.assert_allclose(dynamic_array.data, [p0])
    assert len(dynamic_array) == 1
    assert dynamic_array.capacity == 3

    p1 = np.random.random(3)
    dynamic_array.append(p1)
    npt.assert_allclose(dynamic_array.data, np.vstack((p0, p1)))
    assert len(dynamic_array) == 2
    assert dynamic_array.capacity == 3

    p2 = np.random.random(3)
    dynamic_array.append(p2)
    npt.assert_allclose(dynamic_array.data, np.vstack((p0, p1, p2)))
    assert len(dynamic_array) == 3
    assert dynamic_array.capacity == 3

    p3 = np.random.random(3)
    dynamic_array.append(p3)
    npt.assert_allclose(dynamic_array.data, np.vstack((p0, p1, p2, p3)))
    assert len(dynamic_array) == 4
    assert dynamic_array.capacity == 6



"""Basic class for TreeGrower Algorithms for space colonization."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging

import numpy as np
from numpy.core.umath import clip

from neurots.astrocyte.tmd_utils import barcodes_greater_than_distance
from neurots.astrocyte.tmd_utils import scale_barcode
from neurots.generate.algorithms.common import section_data
from neurots.generate.algorithms.tmdgrower import TMDAlgo
from neurots.morphmath import sample
from neurots.morphmath.bifurcation import directional
from neurots.morphmath.utils import from_to_direction
from neurots.morphmath.utils import in_squared_proximity
from neurots.morphmath.utils import norm as vectorial_norm
from neurots.morphmath.utils import upper_half_ball_query

L = logging.getLogger(__name__)


ENDFOOT_SQUARED_DISTANCE = 5.0


def _majorize_process(process, stop, target_distance):
    """Curates the non-major processes to apply a gradient to large components.

    .. note::
        If the expected maximum length is larger that target distance, then a process type of major
        will be returned. Otherwise, the initial type process will be returned.

    Args:
        process (str): Process type.
        stop (StopCriteria): Section stop criteria.
        target_distance (float): Target distance to check the stop criteria against.

    Returns:
        str: process type
    """
    difference = stop.expected_maximum_length()
    if not np.isinf(difference) and difference > target_distance:
        return "major"
    return process


def _repulsion(points, current_point, length_constant):
    """Calculate the repulsion from the points to the ref_point.

    The length_constant determines the decay rate of the exponential contribution.

    Only the points that are inside the hemisphere with radius cutoff_distance and in the
    same halfspace as section_radius are taken into account.

    Args:
        points (numpy.ndarray): Seed points, the sources of the repulsion force.
        current_point (numpy.ndarray): Reference point on which the repulsion will be applied.
        length_constant (float): Exponential's decay length constant for repulsion strength.

    Returns:
        numpy.ndarray: The average absolute repulsion contribution.
    """
    if len(points) == 0:
        return np.zeros(3, dtype=np.float32)

    vectors = points - current_point

    decay_rate = 1.0 / length_constant
    lengths = np.linalg.norm(vectors, axis=1)
    u_vectors = vectors / lengths[:, None]

    contributions = np.exp(lengths * -decay_rate)
    return contributions.dot(u_vectors) / len(vectors)


def _fallback_strategy(section_direction, angles, repulsion):
    """Directional splitting.

    dir1 is always same as section_direction plus the repulsion.
    dir2 is determined by the angles and the repulsion.

    Args:
        section_direction (numpy.ndarray): The direction of the section.
        angles (list): The list angle distribution to sample from.
        repulsion (numpy.ndarray): The absolute repulsion contribution.

    Returns:
        tuple[numpy.ndarray, numpy.ndarray]: first and second split directions.
    """
    dir1 = section_direction - repulsion
    dir1 /= vectorial_norm(dir1)

    _, dir2 = directional(dir1, angles=angles)

    dir2 -= repulsion
    dir2 /= vectorial_norm(dir2)

    return dir1, dir2


def _colonization_strategy_primary(previous_direction, vectors_to_seeds, repulsion):
    """Compute directions with primary strategy.

    .. note::
        dir1 is calculated by the mean direction from the vectors_to_seeds.

        dir2 is determined by the direction that maximized its angle to dir1.

        Both dir1 and dir2 and dir2 are affected by repulsion.

    Args:
        previous_direction (numpy.ndarray): The previous section direction.
        vectors_to_seeds (numpy.ndarray): Vectors from current point to seeds.
        repulsion (numpy.ndarray): Absolute repulsion contribution.

    Returns:
        tuple[numpy.ndarray, numpy.ndarray]: first and second split directions.
    """
    vectors_to_seeds = vectors_to_seeds - repulsion
    lengths = np.linalg.norm(vectors_to_seeds, axis=1)
    uvecs = vectors_to_seeds / lengths[:, None]

    dir1 = previous_direction - repulsion
    dir1 /= vectorial_norm(dir1)

    # most negative dot product of unit vecs corresponds to the largest angle
    pos2 = np.argmin(uvecs.dot(dir1))
    dir2 = uvecs[pos2]

    return dir1, dir2


def _colonization_strategy_secondary(vectors_to_seeds, repulsion):
    """Compute directions with secondary strategy.

    .. note::
        dir1 is calculated by the mean direction from the vectors_to_seeds.

        dir2 is determined by the direction that maximized its angle to dir1.

        Both dir1 and dir2 and dir2 are affected by repulsion.

    Args:
        vectors_to_seeds (numpy.ndarray): Vectors from current point to seeds.
        repulsion (numpy.ndarray): Absolute repulsion contribution.

    Returns:
        tuple[numpy.ndarray, numpy.ndarray]: The first and second split directions.
    """
    vectors_to_seeds = vectors_to_seeds - repulsion
    lengths = np.linalg.norm(vectors_to_seeds, axis=1)
    uvecs = vectors_to_seeds / lengths[:, None]

    pos1 = np.argmin(lengths)
    dir1 = uvecs[pos1]

    # most negative dot product of unit vecs corresponds to the largest angle
    pos2 = np.argmin(uvecs.dot(dir1))
    dir2 = uvecs[pos2]

    return dir1, dir2


def _colonization_split(section, angles, parameters, context):
    """Creates a bifurcation using the space colonization algorithm.

    Args:
        section (SectionGrower): The current section
        angles (list): Angle distribution for sampling
        parameters (dict): Algorithm's parameters
        context (SpaceColonizationContext): External space colonization data

    Returns:
        tuple[numpy.ndarray, str, numpy.ndarray, str]:
            A tuple containing the following values:

            * first direction
            * first section type
            * second direction
            * second section type

    .. note::
        A repulsion contribution is calculated first from the morphology points
        that are around the bifurcation point. If there are no points available
        the repulsion will be zero.

        If the section grower is a major type:

        A partial ball query is performed to find all the point cloud seeds
        that are around the bifurcation point. The primary colonization strategy
        is then used where the first direction is the same as the direction of the
        section plus the repulsion and the second one is chosen so that it maximized
        the angle with the first direction.

        if the section grower is not a major type:

        A half bal query is perfmored along the direction of the section to find the
        point cloud seeds around the bifurcation point. The secondary colonization
        strategy is then used where the first direction is chosen by the closest seed
        to the bifurcation point and the second one from the direction that maximizes
        the direction to the first one.

        In both cases if not enough seeds are found the strategy will fall back to a
        directional splitting sampling from the bio angles. It will also fallback if
        the two new directions are identical. It's a rare case but it can happen given
        overlapping point cloud seeds.
    """
    current_point = section.last_point
    section_direction = section.direction

    segment_length = parameters["step_size"]["norm"]["mean"]
    kill_distance = context.kill_distance(segment_length)
    influence_distance = context.influence_distance(segment_length)

    point_cloud = context.point_cloud

    # current points of the morphology contributing to self repulsion
    # last point is current_point, therefore we should ignore it
    morphology_points = context.morphology_points.data[:-1]

    # repulsion contribution only from points in the hemisphere aligned to direction
    ids = upper_half_ball_query(morphology_points, current_point, kill_distance, section_direction)
    repulsion = _repulsion(morphology_points[ids], current_point, kill_distance)

    if section.process == "major":

        seed_ids = point_cloud.partial_ball_query(
            current_point,
            influence_distance,
            section_direction,
            np.pi / 3.0,
            np.pi / 2.5,
        )

        if len(seed_ids) < 2:
            dir1, dir2 = _fallback_strategy(section_direction, angles, repulsion)
        else:
            vectors_to_seeds = point_cloud.points[seed_ids] - current_point
            dir1, dir2 = _colonization_strategy_primary(
                section_direction, vectors_to_seeds, repulsion
            )
    else:

        seed_ids = point_cloud.upper_half_ball_query(
            current_point, influence_distance, section_direction
        )

        if len(seed_ids) < 2:
            dir1, dir2 = _fallback_strategy(section_direction, angles, repulsion)
        else:
            vectors_to_seeds = point_cloud.points[seed_ids] - current_point
            dir1, dir2 = _colonization_strategy_secondary(vectors_to_seeds, repulsion)

    if np.allclose(dir1, dir2):

        L.warning(
            "Splitting directions are identical. Use of fallback strategy to recalculate them."
        )
        dir1, dir2 = _fallback_strategy(section_direction, angles, repulsion)

    return dir1, section.process, dir2, "secondary"


def _add_attraction_bias(
    current_point, target_point, max_target_distance, direction, attraction_function
):
    """Adds a bias to the direction whose magnitude depends on the attraction function.

    The attraction function takes as an input the ratio of the distance of the current
    point to the target over the total distance of the target from the soma.

    Args:
        current_point (numpy.ndarray): The reference point.
        target_point (numpy.ndarray): The attractor point.
        max_target_distance (float): The distance from soma center to attraction point.
        direction (numpy.ndarray): Section direction.
        attraction_function (Callable[float] -> float): Attraction strength function handle.

    Returns:
        numpy.ndarray: The new direction.
    """
    target_direction, magnitude = from_to_direction(current_point, target_point, return_length=True)

    fraction = clip(magnitude / max_target_distance, 0.0, 1.0)
    a = clip(attraction_function(fraction), 0.0, 1.0)

    new_direction = a * target_direction + (1.0 - a) * direction
    new_direction /= vectorial_norm(new_direction)

    return new_direction


def _colonization_split_with_target_influence(section, angles, parameters, context):
    """Creates a bifurcation using the space colonization algorithm.

    Args:
        section (SectionGrower): The current section.
        angles (list): Angle distribution for sampling.
        parameters (dict): Algorithm's parameters.
        context (SpaceColonizationContext): External space colonization data.

    Returns:
        tuple[numpy.ndarray, str, numpy.ndarray, str]:
            A tuple containing the following values:

            * first direction
            * first section type
            * second direction
            * second section type

    .. note::
        Bifurcation directions are calculated based on the space colonization split. If
        the endfoot target is active and is close to the bifurcation point, then the the
        second section type will become an endfoot and the target will be exclusive to it,
        rendering it not active. The second direction will be overridden with the direction
        to the target.

        If the target is not in proximity and the section has a major type, an attraction bias
        is added towards the target that depends on the strength of the attraction field. If it
        isn't a major type, it can become major via the majorization function. In order to be
        majorized the bar must be a long one, i.e. longer than 80% of the distance from the tree
        start to the soma.
    """
    process1, process2 = section.process, "secondary"

    current_point = section.last_point
    target_id = section.stop_criteria["target_id"]

    dir1, _, dir2, _ = _colonization_split(section, angles, parameters, context)

    if not context.endfeet_targets.active[target_id]:
        return dir1, process1, dir2, process2

    target_point = context.endfeet_targets.points[target_id]

    if in_squared_proximity(current_point, target_point, ENDFOOT_SQUARED_DISTANCE):

        # this section will become the endfoot section that will connect to the target
        process2 = "endfoot"

        # make the target invisible for the rest of the section growers
        context.endfeet_targets.active[target_id] = False
        dir2 = from_to_direction(current_point, target_point)
    else:

        max_target_distance = parameters["distance_soma_target"]
        if process1 == "major":
            dir1 = _add_attraction_bias(
                current_point,
                target_point,
                max_target_distance,
                dir1,
                section.context.field,
            )
        else:

            process1 = _majorize_process(
                process1,
                section.stop_criteria["TMD"],
                parameters["bias"] * max_target_distance,
            )

    return dir1, process1, dir2, process2


class SpaceColonization(TMDAlgo):
    """Algorithm for growing a tree using a variation of the space colonization algorithm.

    A spatial colonization context is required, which will provide access
    to the seed point cloud for the respective queries needed for the splitting strategies.

    The neurots space colonization algorithm operates using the following data:

    - SpaceColonizationContext
        - Seeds point cloud
        - Space colonization parameters (kill and influence distance)
    """

    def select_persistence(self, input_data, random_generator=np.random):
        """Select the persistence.

        The persistaece is randomly selected from the barcodes the max radial of which
        is greater or equal to the distance from the soma to the domain face.

        Args:
            input_data (dict): The input data parameters.
            random_generator (numpy.random.Generator): The random number generator to use.

        Returns:
            Barcode: The topology barcode.
        """
        if "distance_to_domain" not in self.params:
            return super().select_persistence(input_data, random_generator)

        target_distance = self.params["distance_to_domain"]

        persistence = sample.ph(
            barcodes_greater_than_distance(input_data["persistence_diagram"], target_distance),
            random_generator,
        )

        if self.params["barcode_scaling"]:
            persistence = scale_barcode(persistence, target_distance)

        return persistence

    def bifurcate(self, current_section):
        """When the section bifurcates two new sections need to be created.

        This method computes from the current state the data required for the
        generation of two new sections and returns the corresponding dictionaries.
        """
        self.barcode.remove_bif(current_section.stop_criteria["TMD"].bif_id)
        ang = self.barcode.angles[current_section.stop_criteria["TMD"].bif_id]

        dir1, type1, dir2, type2 = _colonization_split(
            current_section, ang, self.params, self.context
        )

        first_point = np.array(current_section.last_point)

        stop1, stop2 = self.get_stop_criteria(current_section)

        return (
            section_data(dir1, first_point, stop1, type1),
            section_data(dir2, first_point, stop2, type2),
        )


class SpaceColonizationTarget(SpaceColonization):
    """A target is specified for this algorithm.

    The tree grows biased from the target and when it reaches it, it stops being influenced by the
    point.
    """

    def select_persistence(self, input_data, random_generator=np.random):
        """Select the persistence.

        The persistaece is randomly selected from the barcodes the max radial of which is greater
        or equal to the distance from the soma to the target.

        Args:
            input_data (dict): The input data parameters.
            random_generator (numpy.random.Generator): The random number generator to use.

        Returns:
            Barcode: The topology barcode.
        """
        target_distance = self.params["distance_soma_target"]

        persistence = sample.ph(
            barcodes_greater_than_distance(input_data["persistence_diagram"], target_distance),
            random_generator,
        )

        if self.params["barcode_scaling"]:
            persistence = scale_barcode(persistence, target_distance)

        return persistence

    def initialize(self):
        """TMD basic grower of an apical tree based on path distance.

        Initializes the tree grower and computes the apical distance using the input barcode.
        """
        stop_criteria, n_sections = super().initialize()
        stop_criteria["target_id"] = self.params["target_id"]
        return stop_criteria, n_sections

    def get_stop_criteria(self, current_section):
        """Get stop criteria.

        Use the stop criteria of the parent class with the additional information
        of the target_id that is assigned to the tree and the max_target_distance for
        calculating the magnitude of the attraction field generated by the target.
        """
        stop1, stop2 = super().get_stop_criteria(current_section)
        stop1["target_id"] = stop2["target_id"] = self.params["target_id"]
        return stop1, stop2

    def bifurcate(self, current_section):
        """When the section bifurcates two new sections need to be created.

        This method computes from the current state the data required for the
        generation of two new sections and returns the corresponding dictionaries.
        """
        self.barcode.remove_bif(current_section.stop_criteria["TMD"].bif_id)
        ang = self.barcode.angles[current_section.stop_criteria["TMD"].bif_id]

        dir1, type1, dir2, type2 = _colonization_split_with_target_influence(
            current_section, ang, self.params, self.context
        )

        first_point = np.array(current_section.last_point)

        stop1, stop2 = self.get_stop_criteria(current_section)

        return (
            section_data(dir1, first_point, stop1, type1),
            section_data(dir2, first_point, stop2, type2),
        )



"""Test neurots.astrocyte.tree code."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=missing-function-docstring
import numpy as np
from mock import Mock

from neurots.astrocyte.context import SpaceColonizationContext
from neurots.astrocyte.section import SectionSpatialGrower
from neurots.astrocyte.tree import TreeGrowerSpaceColonization

from .test_grower import _context
from .test_grower import _distributions
from .test_grower import _parameters


def test_tree_grower_space_colonization_constructor():

    neuron = Mock()
    initial_direction = np.array([0.12427115, 0.93206836, 0.3403017])
    initial_point = np.array([1.5411615, 0.29389329, 0.15904417])

    grower_distributions = _distributions()
    grower_parameters = _parameters()
    skip_validation = False
    grower_context = SpaceColonizationContext(_context())
    tree_parameters = grower_parameters["basal"]
    tree_parameters["origin"] = np.zeros(3)
    tree_distributions = grower_distributions["basal"]

    tree_grower = TreeGrowerSpaceColonization(
        neuron,
        initial_direction,
        initial_point,
        tree_parameters,
        tree_distributions,
        skip_validation,
        grower_context,
    )

    assert isinstance(tree_grower.active_sections[0], SectionSpatialGrower)

    tree_parameters = grower_parameters["axon"]
    tree_parameters["origin"] = np.zeros(3)
    tree_parameters["target_id"] = 0
    tree_parameters["distance_soma_target"] = 1.2
    tree_distributions = grower_distributions["axon"]

    tree_grower = TreeGrowerSpaceColonization(
        neuron,
        initial_direction,
        initial_point,
        tree_parameters,
        tree_distributions,
        skip_validation,
        grower_context,
    )
    assert isinstance(tree_grower.active_sections[0], SectionSpatialGrower)



"""Test NeuroTS.utils code."""

# Copyright (C) 2022  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=missing-function-docstring
import json
from copy import deepcopy
from pathlib import Path

import dictdiffer  # pylint: disable=import-error
import numpy as np
import pytest

from neurots import utils

DATA = Path(__file__).parent / "data"


def test_format_values():
    data = {
        "array_int": np.array([1, 2]),
        "array_float": np.array([1.1, 2.2]),
        "array_obj": np.array(["a", [1, 2]], dtype=object),
        "2d_array_int": np.array([[1, 2]], ndmin=2),
        "2d_array_float": np.array([[1.1, 2.2]], ndmin=2),
        "2d_array_obj": np.array([["a", [1, 2]]], ndmin=2, dtype=object),
        "float32": np.float32(1.1),
        "float64": np.float64(1.1),
        "int32": np.int32(1),
        "bool8": np.bool8(True),
    }
    data["sub_dict"] = deepcopy(data)
    data["sub_list"] = [deepcopy(data), deepcopy(data)]

    expected = {
        "array_int": [1, 2],
        "array_float": [1.1, 2.2],
        "array_obj": ["a", [1, 2]],
        "2d_array_int": [[1, 2]],
        "2d_array_float": [[1.1, 2.2]],
        "2d_array_obj": [["a", [1, 2]]],
        "float32": 1.1,
        "float64": 1.1,
        "int32": 1,
        "bool8": True,
    }
    expected["sub_dict"] = deepcopy(expected)
    expected["sub_list"] = [deepcopy(expected), deepcopy(expected)]

    with pytest.raises(AssertionError):
        assert not list(dictdiffer.diff(utils.format_values(data), expected))
    assert not list(dictdiffer.diff(utils.format_values(data, decimals=6), expected))


def test_convert_from_legacy_neurite_type():
    """Test convert legacy json files (with added _dendrite to basal/apical)."""

    with open(DATA / "dummy_distribution.json", encoding="utf-8") as f:
        data = json.load(f)

    data_converted = utils.convert_from_legacy_neurite_type(data)
    assert data_converted == data

    with open(DATA / "dummy_distribution_legacy.json", encoding="utf-8") as f:
        data_legacy = json.load(f)

    data_converted = utils.convert_from_legacy_neurite_type(data_legacy)
    assert data_converted == data

    with open(DATA / "dummy_params.json", encoding="utf-8") as f:
        data = json.load(f)

    data_converted = utils.convert_from_legacy_neurite_type(data)
    assert data_converted == data

    with open(DATA / "dummy_params_legacy.json", encoding="utf-8") as f:
        data_legacy = json.load(f)

    data_converted = utils.convert_from_legacy_neurite_type(data_legacy)
    assert data_converted == data



# noqa

# Copyright (C) 2022  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Extract inputs for synthesis
============================

This example shows how to extract the inputs required for synthesis from a set of existing
morphologies.
"""

import json
from pathlib import Path

import neurots
from neurots import extract_input


def run(output_dir, data_dir):
    """Run the example for extracting inputs for synthesis."""
    # Generate distribution from directory of neurons
    distr = extract_input.distributions(
        data_dir / "neurons", feature="path_distances", diameter_model="default"
    )

    # Save distributions in a json file
    with open(output_dir / "test_distr.json", "w", encoding="utf-8") as f:
        json.dump(distr, f, sort_keys=True, indent=2)

    # Generate default parameters for topological synthesis of basal dendrites
    params = extract_input.parameters(
        neurite_types=["basal_dendrite"], feature="path_distances", method="tmd"
    )

    # Save parameters in a json file
    with open(output_dir / "test_params.json", "w", encoding="utf-8") as f:
        json.dump(params, f, sort_keys=True, indent=2)

    # Re-load data from saved distributions
    with open(output_dir / "test_distr.json", "r", encoding="utf-8") as F:
        distr = json.load(F)

    # Re-load data from saved parameters
    with open(output_dir / "test_params.json", "r", encoding="utf-8") as F:
        params = json.load(F)

    # Initialize a neuron
    N = neurots.NeuronGrower(input_distributions=distr, input_parameters=params)

    # Grow the neuron
    neuron = N.grow()

    # Export the synthesized cell
    neuron.write(output_dir / "generated_cell.asc")
    neuron.write(output_dir / "generated_cell.swc")
    neuron.write(output_dir / "generated_cell.h5")


if __name__ == "__main__":
    result_dir = Path("results_extract_synthesis_inputs")
    result_dir.mkdir(parents=True, exist_ok=True)

    run(result_dir, Path("data"))



"""Abstract class for TreeGrower Algorithms."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import abc
import copy


class AbstractAlgo:
    """TreeGrower abstract class.

    Args:
        input_data (dict): All the data required for the growth.
        params (dict): The parameters required for growth.
        start_point (list[float]): The first point of the trunk.
        context (Any): An object containing contextual information.
    """

    # meta class is used to define other classes
    __metaclass__ = abc.ABCMeta

    def __init__(self, input_data, params, start_point, context):
        """The TreeGrower Algorithm initialization."""
        self.context = context
        self.input_data = copy.deepcopy(input_data)
        self.params = copy.deepcopy(params)
        self.start_point = start_point

    @abc.abstractmethod
    def initialize(self):
        """Abstract TreeGrower Algorithm initialization.

        Generates the data to be used for the initialization of the first section to be grown.
        """

    @abc.abstractmethod
    def bifurcate(self, current_section):
        """When the section bifurcates two new sections need to be created.

        This method computes from the current state the data required for the
        generation of two new sections and returns the corresponding dictionaries.
        """

    @abc.abstractmethod
    def terminate(self, current_section):
        """Terminate the current section.

        When the growth of a section is terminated the "term" must be removed from the TMD grower.
        """

    @abc.abstractmethod
    def extend(self, current_section):
        """Definition of stop criterion for the growth of the current section."""



"""Test neurots.generate.section code."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json
import os

import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal
from numpy.testing import assert_array_equal
from numpy.testing import assert_equal

from neurots.generate.algorithms.barcode import Barcode
from neurots.generate.algorithms.common import TMDStop
from neurots.generate.algorithms.common import checks_bif_term
from neurots.utils import NeuroTSError

_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")


def test_barcode():
    """Tests the barcode functionality"""

    with open(os.path.join(_PATH, "dummy_distribution.json"), encoding="utf-8") as f:
        ph_angles = json.load(f)["apical_dendrite"]["persistence_diagram"][0]

    barcode_test = Barcode(ph_angles)
    assert_equal(barcode_test.get_term_between(999), (None, -np.inf))
    assert_array_equal(barcode_test.min_bif(np.inf), (1, 26.3027))
    assert_array_equal(barcode_test.min_term(np.inf), (2, 155.8809))

    assert_equal(len(barcode_test.bifs), 7)
    assert_equal(len(barcode_test.terms), 8)
    assert_equal(len(barcode_test.angles), 8)
    assert_array_almost_equal(barcode_test.angles[0], [np.nan, np.nan, np.nan, np.nan])

    assert_equal(barcode_test.bifs[1], 26.3027)
    assert_equal(barcode_test.terms[1], 204.0442)
    assert_array_almost_equal(barcode_test.angles[1], [-0.279172, -0.138322, 1.598153, 0.672246])

    assert_array_almost_equal(barcode_test.get_bar(0), (0, 633.5966))
    assert_array_almost_equal(barcode_test.get_bar(1), (26.3027, 204.0442))

    assert_array_almost_equal(barcode_test.min_bif(), (1, 26.3027))
    assert_array_almost_equal(barcode_test.min_term(), (2, 155.8809))
    assert_array_almost_equal(barcode_test.max_term(), (0, 633.5966))

    barcode_test.remove_bif(1)
    assert_array_almost_equal(barcode_test.min_bif(), (2, 52.4013))

    barcode_test.remove_term(2)
    assert_array_almost_equal(barcode_test.min_term(), (1, 204.0442))


def test_barcode_validate_persistence():
    """Tests the barcode functionality"""

    with open(os.path.join(_PATH, "dummy_distribution.json"), encoding="utf-8") as f:
        ph_angles = json.load(f)["apical_dendrite"]["persistence_diagram"][0]
    assert Barcode.validate_persistence(ph_angles)
    ph_angles[0][0] = 0
    assert not Barcode.validate_persistence(ph_angles)


def test_checks_bif_term():
    """Test the checks_bif_term() function"""
    assert checks_bif_term(0, 1, 2, 10)
    assert not checks_bif_term(2, 1, 0, 10)

    assert checks_bif_term(0, np.inf, 2, 10)
    assert not checks_bif_term(2, np.inf, 0, 10)

    assert not checks_bif_term(0, 1, 2, 0.5)
    assert not checks_bif_term(0, np.inf, 2, 0.5)

    assert checks_bif_term(0, 1, 2, 2)
    assert not checks_bif_term(2, 1, 0, 2)


def test_TMDStop():
    """Test the TMDStop class"""
    tmd_stop = TMDStop(1, 26.3027, 0, 633.5966, 10.0)
    assert str(tmd_stop) == "(Ref: 10.0, BifID: 1, Bif: 26.3027, TermID: 0, Term: 633.5966)"
    assert_equal(tmd_stop.bif_id, 1)
    assert_equal(tmd_stop.term_id, 0)
    assert_equal(tmd_stop.bif, 26.3027)
    assert_equal(tmd_stop.term, 633.5966)
    assert_equal(tmd_stop.ref, 10.0)
    assert_equal(tmd_stop.child_length(), 607.2939)
    assert_equal(tmd_stop.expected_termination_length(), 623.5966)
    # Comparison should pass for same TMDstop
    assert tmd_stop == TMDStop(1, 26.3027, 0, 633.5966, 10.0)
    # Comparison should fail for not same TMDstop
    assert tmd_stop != TMDStop(1, 20, 0, 633.5966, 10.0)
    # Comparison should fail for other object
    assert tmd_stop != 10

    tmd_stop.update_bif(0, 11.0)
    assert_equal(tmd_stop.bif_id, 0)
    assert_equal(tmd_stop.bif, 11.0)

    tmd_stop.update_term(10, 110.0)
    assert_equal(tmd_stop.term_id, 10)
    assert_equal(tmd_stop.term, 110.0)

    assert tmd_stop.verify()
    assert_equal(tmd_stop.expected_bifurcation_length(), 1)

    tmd_stop.bif = np.inf
    assert tmd_stop.verify()
    assert_equal(tmd_stop.expected_bifurcation_length(), 0)

    tmd_stop.ref = 99999
    assert not tmd_stop.verify()

    tmd_stop.term = np.inf
    assert_equal(tmd_stop.expected_termination_length(), 0)

    with open(os.path.join(_PATH, "dummy_distribution.json"), encoding="utf-8") as f:
        ph_angles = json.load(f)["apical_dendrite"]["persistence_diagram"][0]

    barcode_test = Barcode(ph_angles)

    parent_stop = TMDStop(1, 26.3027, 0, 10, 10.0)
    child_stop = TMDStop(1, 26.3027, 0, 999999, 10.0)
    with pytest.raises(NeuroTSError):
        barcode_test.curate_stop_criterion(parent_stop, child_stop)

    child_stop = TMDStop(1, 26.3027, 999999, 5, 10.0)
    with pytest.raises(NeuroTSError):
        barcode_test.curate_stop_criterion(parent_stop, child_stop)



"""NeuroTS class: Grower object that contains the grower functionality."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
from copy import deepcopy

import numpy as np
from morphio import SectionType
from scipy.spatial import KDTree

from neurots.astrocyte.context import SpaceColonizationContext
from neurots.astrocyte.section import grow_to_target
from neurots.astrocyte.tree import TreeGrowerSpaceColonization
from neurots.generate.grower import NeuronGrower
from neurots.morphmath import sample
from neurots.morphmath.utils import norm as vectorial_norm
from neurots.morphmath.utils import normalize_vectors

L = logging.getLogger(__name__)


def _number_of_trees(tree_type, oris, distributions, random_generator=np.random):
    """Sample the number of trees depending on the tree type if no predef orientations."""
    if oris is None:
        n_trees = sample.n_neurites(distributions["num_trees"], random_generator)
    else:
        n_trees = len(oris)

    if tree_type == "basal" and n_trees < 2:
        raise Exception(f"There should be at least 2 basal dendrites (got {n_trees})")

    return n_trees


def _ensure_endfeet_are_reached(cell, targets):
    """Ensure Endfeet points are reached.

    If there are any targets left active, i.e. not reached, the closest section to each one
    of them will be grown to reach it. This is the last resort to ensure that endfeet targets will
    be reached no matter what. However, a warning will be emitted.
    """
    perivascular_terminations = [
        s for s in cell.iter() if s.type == SectionType(2) and len(s.children) == 0
    ]

    termination_points = np.array([s.points[-1] for s in perivascular_terminations])

    target_points = targets.active_points
    distances, section_indices = KDTree(termination_points, copy_data=False).query(target_points)

    for distance, section_index, target_point in zip(distances, section_indices, target_points):

        if not np.isclose(distance, 0.0):

            L.warning(
                "Target %s was not reached. Extending closest section to reach it.",
                target_point,
            )

            section = perivascular_terminations[section_index]
            points, diameters = section.points, section.diameters

            vector = points[-1] - points[-2]
            segment_length = np.linalg.norm(vector)
            direction = vector / segment_length

            new_points = grow_to_target(points[-1], direction, target_point, segment_length)

            section.points = np.vstack((points, new_points))
            section.diameters = np.hstack(
                (diameters, np.full(len(new_points), fill_value=diameters[-1]))
            )


class AstrocyteGrower(NeuronGrower):
    """The specific grower for astrocytes.

    A Grower object is a container for an Astrocyte, encoded in the (groups, points) structure,
    as a morphIO Morphology object. A set of input distributions that store the data
    consumed by the algorithms and the user-selected parameters are also stored.
    """

    def __init__(
        self,
        input_parameters,
        input_distributions,
        context,
        external_diametrizer=None,
        skip_validation=False,
        rng_or_seed=np.random,
    ):
        super().__init__(
            input_parameters,
            input_distributions,
            context=SpaceColonizationContext(context),
            external_diametrizer=external_diametrizer,
            skip_validation=skip_validation,
            rng_or_seed=rng_or_seed,
        )

    def validate_params(self):
        """Astrocyte parameter validation."""

    def validate_distribs(self):
        """Astrocyte distribution validation."""

    def _post_grow(self):
        """After the cell has grown ensure that endfeet targets have been reached."""
        endfeet_targets = self.context.endfeet_targets
        if endfeet_targets is not None and np.any(endfeet_targets.active):
            _ensure_endfeet_are_reached(self.neuron, self.context.endfeet_targets)

    def _add_active_neurite(self, initial_soma_point, parameters, distributions):
        """Create a TreeGrower with the input data and add it to the active neurites."""
        tree_direction = self.soma_grower.soma.orientation_from_point(initial_soma_point)

        obj = TreeGrowerSpaceColonization(
            self.neuron,
            initial_direction=tree_direction,
            initial_point=initial_soma_point,
            parameters=parameters,
            distributions=distributions,
            context=self.context,
            random_generator=self._rng,
        )

        self.active_neurites.append(obj)

    def _orientations_to_points(self, oris, n_trees, distr):
        """Return soma points from directions.

        Checks the type of orientation input and returns the soma points generated by the
        corresponding selection.

        Currently accepted orientations include the following options:
            * list of 3D points: select the orientation externally
            * None: creates a list of orientations according to the biological distributions.
            * 'from_space': generates orientations depending on spatial input (not implemented yet).
        """
        if oris is None:
            assert n_trees != 0, "Number of trees should be greater than zero"
            trunk_angles = sample.trunk_angles(distr, n_trees, self._rng)
            trunk_z = sample.azimuth_angles(distr, n_trees, self._rng)
            return self.soma_grower.add_points_from_trunk_angles(trunk_angles, trunk_z)

        assert len(oris) >= n_trees, "n_orientations < n_trees"
        return self.soma_grower.add_points_from_orientations(oris[:n_trees])

    def _create_process_trunks(self, tree_type, parameters, distributions):
        """Creates the starts of the processes for the astrocyte.

        Args:
            tree_type (str)
            parameters (dict)
            distributions (distributions)

        Notes:
            Astrocytic perivascular type corresponds to the neuronal axon, while
            perisynaptic type to basal or apical type.
        """
        tree_type = "perivascular" if tree_type == "axon" else "perisynaptic"

        if tree_type == "perivascular":

            target_ids = np.array(parameters["target_ids"], dtype=np.int32)
            target_points = self.context.endfeet_targets.points[target_ids]

            oris = normalize_vectors(
                np.array(
                    [self.soma_grower.soma.orientation_from_point(p) for p in target_points],
                    dtype=np.float32,
                )
            )

            trunk_points = self.soma_grower.add_points_from_orientations(oris)

            assert len(target_ids) == len(
                trunk_points
            ), "Number of targets is not equal to number of orientations"

            for target_id, trunk_point in zip(target_ids, trunk_points):

                neurite_params = deepcopy(parameters)
                neurite_params["target_id"] = target_id
                neurite_params["distance_soma_target"] = vectorial_norm(
                    self.soma_grower.soma.center - self.context.endfeet_targets.points[target_id]
                )
                self._add_active_neurite(trunk_point, neurite_params, distributions)
        else:

            oris = parameters["orientation"]

            if oris is not None:
                oris = normalize_vectors(np.array(oris, dtype=np.float32))

            trunk_points = self._orientations_to_points(
                oris,
                _number_of_trees(tree_type, oris, distributions, self._rng),
                distributions,
            )

            for i, trunk_point in enumerate(trunk_points):

                neurite_params = deepcopy(parameters)
                if "domain_distances" in neurite_params:
                    neurite_params["distance_to_domain"] = parameters["domain_distances"][i]
                self._add_active_neurite(trunk_point, neurite_params, distributions)

    def _grow_trunks(self):
        """Grow the trunks.

        Generates the initial points of each tree, which depend on the selected
        tree types and the soma surface. All the trees start growing from the surface
        of the soma. The outgrowth direction is either specified in the input parameters,
        as parameters['type']['orientation'] or it is randomly chosen according to the
        biological distribution of trunks on the soma surface if 'orientation' is None.
        """
        origin = np.array(self.input_parameters["origin"], dtype=np.float32)

        for tree_type in self.input_parameters["grow_types"]:
            parameters = self.input_parameters[tree_type]
            distributions = self.input_distributions[tree_type]
            parameters["origin"] = origin
            self._create_process_trunks(tree_type, parameters, distributions)



"""Test neurots.generate.diametrizer code."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=missing-function-docstring
# pylint: disable=no-member
# pylint: disable=redefined-outer-name
import copy
import os

import morphio
import numpy as np
import pytest
from morphio import SectionType
from numpy.testing import assert_array_almost_equal
from numpy.testing import assert_equal

from neurots.generate import diametrizer

_path = os.path.dirname(os.path.abspath(__file__))
NEU_PATH1 = os.path.join(_path, "../test_data/diam_simple.swc")
NEU_PATH2 = os.path.join(_path, "../test_data/simple.swc")
NEU_PATH3 = os.path.join(_path, "../test_data/diam_simple_axon.swc")

MODEL = {
    "basal_dendrite": {
        "Rall_ratio": 2.0 / 3.0,
        "siblings_ratio": 1.0,
        "taper": [0.1],
        "term": [0.6],
        "trunk": [4.0, 3.0],
        "trunk_taper": [0.6],
    },
    "axon": {
        "Rall_ratio": 2.0 / 3.0,
        "siblings_ratio": 1.0,
        "taper": [0.1],
        "term": [0.6],
        "trunk": [4.0, 3.0],
        "trunk_taper": [0.6],
    },
}


@pytest.fixture
def neu1():
    return morphio.mut.Morphology(NEU_PATH1)


@pytest.fixture
def neu2():
    return morphio.mut.Morphology(NEU_PATH2)


@pytest.fixture
def neu3():
    return morphio.mut.Morphology(NEU_PATH3)


def test_sample():
    assert diametrizer.sample([0.0]) == 0.0
    np.random.seed(0)
    assert diametrizer.sample([1.0, 1.0, 1.0, 2.0, 2.0, 3.0]) == 2.0


def test_bifurcator():
    d1 = diametrizer.bifurcator(1.0, 2.0, 3.0 / 2.0, 1.0)
    assert d1 == 0.6299605249474366
    d1 = diametrizer.bifurcator(1.0, 2.0, 1.0, 1.0)
    assert d1 == 0.5
    d1 = diametrizer.bifurcator(1.0, 2.0, 1.0, 0.5)
    assert d1 == 0.6666666666666666


def test_taper_section_diam_from_root(neu1):
    section = neu1.root_sections[0]

    diametrizer.taper_section_diam_from_root(section, 4.0, 0.6, 0.07, 100.0)

    assert_array_almost_equal(section.diameters, [4.0, 3.9, 3.8, 3.7, 3.6, 3.5, 3.4])

    section1 = section.children[0]
    diametrizer.taper_section_diam_from_root(section1, 4.0, 0.5, 0.07, 100.0)
    assert_array_almost_equal(section1.diameters, [4.0, 4.0, 3.9, 3.8, 3.7, 3.6])

    section2 = section.children[0]
    diametrizer.taper_section_diam_from_root(section2, 4.0, 0.5, 99, 100.0)
    assert_array_almost_equal(section2.diameters, [4, 99, 99, 99, 99, 99])


def test_taper_section_diam_from_tips(neu1):
    section = neu1.root_sections[0]

    diametrizer.taper_section_diam_from_tips(section, 4.0, 0.6, 0.07, 100.0)

    assert_array_almost_equal(section.diameters, [4.6, 4.5, 4.4, 4.3, 4.2, 4.1, 4.0])

    section1 = section.children[0]
    diametrizer.taper_section_diam_from_tips(section1, 4.0, 0.5, 0.07, 100.0)
    assert_array_almost_equal(section1.diameters, [4.5, 4.4, 4.3, 4.2, 4.1, 4.0])

    section2 = section.children[0]
    diametrizer.taper_section_diam_from_tips(section2, 4.0, 0.5, 99, 99.2)
    assert_array_almost_equal(section2.diameters, [99.2, 99.2, 99.2, 99.1, 99, 4.0], decimal=5)


def test_diametrize_constant_per_section(neu2):
    diametrizer.diametrize_constant_per_section(neu2)
    assert_array_almost_equal(
        morphio.Morphology(neu2).diameters,
        [2.0, 2.0, 2.5, 2.5, 2.5, 2.5, 2.0, 2.0, 3.0, 3.0, 3.0, 3.0],
    )


def test_diametrize_constant_per_neurite(neu2):
    diametrizer.diametrize_constant_per_neurite(neu2)
    assert_array_almost_equal(
        morphio.Morphology(neu2).diameters,
        [
            2.333333,
            2.333333,
            2.333333,
            2.333333,
            2.333333,
            2.333333,
            2.666667,
            2.666667,
            2.666667,
            2.666667,
            2.666667,
            2.666667,
        ],
    )


def test_diametrize_smoothing(neu1):
    diametrizer.diametrize_smoothing(neu1)
    assert_array_almost_equal(
        morphio.Morphology(neu1).diameters,
        [
            4.0,
            3.9,
            3.8,
            3.7,
            3.6,
            3.5,
            3.4,
            2.8,
            2.8,
            2.704,
            2.608,
            2.512,
            2.416,
            2.8,
            2.8,
            2.76,
            2.72,
            2.68,
            2.64,
        ],
    )


def test_diametrize_from_root(neu1):
    np.random.seed(0)  # ensure constant random number for sampling
    diametrizer.diametrize_from_root(neu1, model_params=MODEL)
    assert_array_almost_equal(
        morphio.Morphology(neu1).diameters,
        [
            4.0,
            3.9,
            3.8,
            3.7,
            3.6,
            3.5,
            3.4,
            3.4,
            1.202082,
            1.182082,
            1.162082,
            1.142082,
            1.122082,
            3.4,
            1.202082,
            1.182082,
            1.162082,
            1.142082,
            1.122082,
        ],
    )


def test_diametrize_from_root_axon(neu3):
    np.random.seed(0)  # ensure constant random number for sampling
    diametrizer.diametrize_from_root(neu3, SectionType.axon, model_params=MODEL)
    assert_array_almost_equal(
        morphio.Morphology(neu3).diameters,
        [
            4.0,
            3.8,
            3.6,
            3.4,
            3.2,
            3.0,
            2.8,
            2.8,
            2.8,
            2.6,
            2.4,
            2.2,
            2.0,
            2.8,
            2.0,
            2.4,
            2.0,
            2.2,
            2.0,
            4.0,
            3.9142857,
            3.8285713,
            3.7428572,
            3.6571429,
            3.5714285,
            3.4857142,
            3.4,
        ],
    )


def test_diametrize_from_tips(neu1):
    np.random.seed(0)  # ensure constant random number for sampling
    diametrizer.diametrize_from_tips(neu1, model_params=MODEL)
    assert_array_almost_equal(
        morphio.Morphology(neu1).diameters,
        [
            2.52333,
            2.423331,
            2.32333,
            2.22333,
            2.12333,
            2.02333,
            1.92333,
            1.92333,
            0.68,
            0.66,
            0.64,
            0.62,
            0.6,
            1.92333,
            0.68,
            0.66,
            0.64,
            0.62,
            0.6,
        ],
    )


def test_diametrize_from_tips_axon(neu3):
    np.random.seed(0)  # ensure constant random number for sampling
    diametrizer.diametrize_from_tips(neu3, model_params=MODEL, neurite_type=SectionType.axon)
    assert_array_almost_equal(
        morphio.Morphology(neu3).diameters,
        [
            4.0,
            3.8,
            3.6,
            3.4,
            3.2,
            3.0,
            2.8,
            2.8,
            2.8,
            2.6,
            2.4,
            2.2,
            2.0,
            2.8,
            2.0,
            2.4,
            2.0,
            2.2,
            2.0,
            1.2,
            1.1142857,
            1.0285715,
            0.94285715,
            0.85714287,
            0.7714286,
            0.6857143,
            0.6,
        ],
    )


def test_redefine_diameter_section(neu1):
    section = neu1.root_sections[0]
    assert_equal(section.diameters[0], 4)
    diametrizer.redefine_diameter_section(section, 0, 999)
    assert_equal(section.diameters[0], 999)

    section.points = np.array([[0, 1, 2]])
    with pytest.raises(Exception):
        diametrizer.redefine_diameter_section(section, 1, 2)


def test_build_M1(neu1):
    diametrizer.build(neu1, diam_method="M1")
    diameters = [i.diameters for i in neu1.sections.values()]
    assert_array_almost_equal(diameters[0], [2.7368422] * 7)
    assert_array_almost_equal(diameters[1], diameters[2])
    assert_array_almost_equal(diameters[2], [2.7368422] * 6)


def test_build_M2(neu1):
    diametrizer.build(neu1, diam_method="M2")
    diameters = [i.diameters for i in neu1.sections.values()]
    assert_array_almost_equal(diameters[0], [3.4] * 7)
    assert_array_almost_equal(diameters[1], [2.4666667] * 6)
    assert_array_almost_equal(diameters[2], [2.2333333] * 6)


def test_build_M3(neu1):
    diametrizer.build(neu1, diam_method="M3")
    diameters = [i.diameters for i in neu1.sections.values()]
    assert_array_almost_equal(diameters[0], [4, 3.9, 3.8, 3.7, 3.6, 3.5, 3.4])
    assert_array_almost_equal(diameters[1], [2.8, 2.8, 2.704, 2.608, 2.512, 2.416])
    assert_array_almost_equal(diameters[2], [2.8, 2.8, 2.76, 2.72, 2.68, 2.64])


def test_build_M4(neu1):
    np.random.seed(1)  # ensure constant random number for sampling
    diametrizer.build(neu1, input_model=MODEL, diam_method="M4")
    diameters = [i.diameters for i in neu1.sections.values()]
    assert_array_almost_equal(diameters[0], [3, 2.9, 2.8, 2.7, 2.6, 2.5, 2.4])
    assert_array_almost_equal(diameters[1], diameters[2])
    assert_array_almost_equal(
        diameters[2], [2.4, 0.84852815, 0.82852817, 0.8085281, 0.78852814, 0.76852816]
    )


def test_build_M5(neu1):
    np.random.seed(0)  # ensure constant random number for sampling
    diametrizer.build(neu1, input_model=MODEL, diam_method="M5")
    diameters = [i.diameters for i in neu1.sections.values()]
    assert_array_almost_equal(
        diameters[0],
        [2.5233305, 2.4233305, 2.3233304, 2.2233305, 2.1233304, 2.0233305, 1.923330],
    )
    assert_array_almost_equal(diameters[1], diameters[2])
    assert_array_almost_equal(diameters[2], [1.9233304, 0.68, 0.66, 0.64, 0.62, 0.6])


def test_build_diam_method(neu1):
    def diam_method(neuron, tree_type, **kwargs):
        # pylint: disable=unused-argument
        diametrizer.diametrize_constant_per_neurite(neuron)

    diametrizer.build(neu1, diam_method=diam_method)
    diameters = [i.diameters for i in neu1.sections.values()]
    assert_array_almost_equal(
        diameters[0],
        [2.7368424, 2.7368424, 2.7368424, 2.7368424, 2.7368424, 2.7368424, 2.7368424],
    )
    assert_array_almost_equal(diameters[1], diameters[2])
    assert_array_almost_equal(
        diameters[2], [2.7368424, 2.7368424, 2.7368424, 2.7368424, 2.7368424, 2.7368424]
    )


def test_build_M1_basal_axon(neu2):
    diametrizer.build(neu2, diam_method="M1", neurite_types=["basal_dendrite", "axon"])
    diameters = [i.diameters for i in neu2.sections.values()]
    assert_array_almost_equal(diameters[0], [2.3333333, 2.3333333])
    assert_array_almost_equal(diameters[1], diameters[2])
    assert_array_almost_equal(diameters[3], [2.6666667, 2.6666667])
    assert_array_almost_equal(diameters[3], diameters[4])
    assert_array_almost_equal(diameters[4], diameters[5])


def test_build_M1_axon(neu2):
    diametrizer.build(neu2, diam_method="M1", neurite_types=["axon"])
    diameters = [i.diameters for i in neu2.sections.values()]
    assert_array_almost_equal(diameters[0], [2.0, 2.0])
    assert_array_almost_equal(diameters[1], [2.0, 3.0])
    assert_array_almost_equal(diameters[1], diameters[2])
    assert_array_almost_equal(diameters[3], [2.6666667, 2.6666667])
    assert_array_almost_equal(diameters[3], diameters[4])
    assert_array_almost_equal(diameters[4], diameters[5])


def test_build_M4_rng(neu1):
    # Test with custom random generator
    test_model = copy.deepcopy(MODEL)
    test_model["basal_dendrite"]["trunk_taper"] = np.arange(0, 10, 0.3)
    diametrizer.build(
        neu1,
        input_model=test_model,
        diam_method="M4",
        random_generator=np.random.default_rng(3),
    )
    diameters = [i.diameters for i in neu1.sections.values()]
    assert_array_almost_equal(diameters[0], [3, 2.9, 2.8, 2.7, 2.6, 2.5, 2.4])
    assert_array_almost_equal(diameters[1], diameters[2])
    assert_array_almost_equal(
        diameters[2], [2.4, 0.84852815, 0.82852817, 0.8085281, 0.78852814, 0.76852816]
    )


def test_build_unknown_model(neu1):
    with pytest.raises(
        KeyError,
        match=(
            "The name of the diametrization method is unknown: 'UNKNOWN' is not in "
            r"\['M1', 'M2', 'M3', 'M4', 'M5'\]"
        ),
    ):
        diametrizer.build(neu1, None, None, "UNKNOWN")


def test_build_empty_model(neu1):
    with pytest.raises(ValueError):
        diametrizer.build(neu1, None, None, object())



"""Test neurots.astrocyte.context code."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=missing-function-docstring
# pylint: disable=protected-access
import numpy as np
from numpy import testing as npt

from neurots.astrocyte import context as tested


def _input_params():
    return {
        "endfeet_targets": np.array([[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]]),
        "space_colonization": {
            "point_cloud": np.array([[2.0, 0.0, 0.0], [2.0, 1.0, 1.0]]),
            "kill_distance_factor": 15.0,
            "influence_distance_factor": 2.0,
        },
        "field": {"type": "logit", "slope": 1.5, "intercept": 0.1},
        "collision_handle": "collision_handle",
    }


def test_constructor():

    c = tested.SpaceColonizationContext(_input_params())

    npt.assert_equal(len(c.endfeet_targets.points), 2)
    npt.assert_equal(len(c.point_cloud.points), 2)

    assert c.collision_handle == "collision_handle"

    npt.assert_equal(c._params["kill_distance_factor"], 15.0)
    npt.assert_equal(c._params["influence_distance_factor"], 2.0)

    npt.assert_equal(len(c.morphology_points), 0)


def test_methods():

    c = tested.SpaceColonizationContext(_input_params())

    npt.assert_allclose(c.kill_distance(0.1), 15.0 * 0.1)
    npt.assert_allclose(c.influence_distance(0.1), 2.0 * 0.1)
